"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


@author: Fabian A. Stamm
"""

import numpy as np
from matplotlib import pyplot as plt
from skimage import measure
from scipy.spatial import distance

def get_gradient_minima(geo_data, GX,GY,GZ=np.nan, direction='z', ref='x'):
    """
        Compute the shared minima of two gradient fields in 3D space.
        The positions of minima are returned as surface vertices.

        Args:
            geo_data (:class:`gempy.data_management.InputData`)
            GX (ndarray): Gradient field in x-direction.
            GY (ndarray): Gradient field in y-direction.
            GZ (ndarray, optional): Gradient field in z-direction. Only needed if
                minima are relative to a different direction than 'z'.
            direction (str, optional, default = 'z'): Direction to which the gradients refer to,
                and to which the minima are related.
                Default for standard topography is 'z'.
            ref (str, optional, default='x'): Reference direction for
                marching cubes to attain vertices. Default is 'x'.
                'x' or 'y' should suffice and lead to the same results
                in most cases.
                If needed, ref can be declared as the 'mean' of both,
                but this comes at high computational costs and is
                not recommended in most cases.

        Returns:
            gradient minima vertices (coordinates in 3D space)
        """
    # for scaling from voxels up to original scale using the original extent
    vox_size_x = np.abs(geo_data.extent[1]-geo_data.extent[0]) / geo_data.resolution[0]
    vox_size_y = np.abs(geo_data.extent[3]-geo_data.extent[2]) / geo_data.resolution[1]
    vox_size_z = np.abs(geo_data.extent[5]-geo_data.extent[4]) / geo_data.resolution[2]
    vox_size_diag = np.sqrt(vox_size_x ** 2 + vox_size_y ** 2 + vox_size_z ** 2)

    if direction == 'z':
        gx = GX
        gy = GY
    elif direction == 'y':
        gx = GY
        gy = GZ
    elif direction == 'x':
        gx = GZ
        gy = GY
    else:
        raise AttributeError(str(direction) + "must be a cartesian direction, i.e. xyz")

    if ref == 'x':
        v_gx0 = measure.marching_cubes_lewiner(gx, 0)[0]
        v_gx = v_gx0
        v_gx[:, 0] = (v_gx0[:, 0] * vox_size_x)+geo_data.extent[0]
        v_gx[:, 1] = (v_gx0[:, 1] * vox_size_y)+geo_data.extent[2]
        v_gx[:, 2] = (v_gx0[:, 2] * vox_size_z)+geo_data.extent[4]
        return v_gx

    elif ref == 'y':
        v_gy0 = measure.marching_cubes_lewiner(gy, 0)[0]
        v_gy = v_gy0
        v_gy[:, 0] = (v_gy0[:, 0] * vox_size_x)+geo_data.extent[0]
        v_gy[:, 1] = (v_gy0[:, 1] * vox_size_y)+geo_data.extent[2]
        v_gy[:, 2] = (v_gy0[:, 2] * vox_size_z)+geo_data.extent[4]
        return v_gy

    elif ref == 'mean':
        # using marching cubes to aquire surfaces (vertices, simplices) that align
        # with the occurrence of zeros of the gradients (gradient minima)
        v_gx0 = measure.marching_cubes_lewiner(gx, 0)[0]
        v_gy0 = measure.marching_cubes_lewiner(gy, 0)[0]

        v_gx = v_gx0
        v_gx[:, 0] = (v_gx0[:, 0] * vox_size_x) + geo_data.extent[0]
        v_gx[:, 1] = (v_gx0[:, 1] * vox_size_y) + geo_data.extent[2]
        v_gx[:, 2] = (v_gx0[:, 2] * vox_size_z) + geo_data.extent[4]

        v_gy = v_gy0
        v_gy[:, 0] = (v_gy0[:, 0] * vox_size_x) + geo_data.extent[0]
        v_gy[:, 1] = (v_gy0[:, 1] * vox_size_y) + geo_data.extent[2]
        v_gy[:, 2] = (v_gy0[:, 2] * vox_size_z) + geo_data.extent[4]

        # calculating this distance takes very long
        # how cut we possibly simplify this?
        dist_gxy = distance.cdist(v_gx, v_gy, 'euclidean')

        # get distance minima and minima positions for both vertices groups
        # this way we can pair 2 vertices from gx and gy based on their
        # common distance which is to be minimal (smaller than to all other points)
        minx = np.min(dist_gxy, axis=1)
        miny = np.min(dist_gxy, axis=0)
        minx_pos = np.argmin(dist_gxy, axis=1)
        miny_pos = np.argmin(dist_gxy, axis=0)

        # set a cut-off value for minimal distance (here: 3D-diagonal of a voxel/2)
        gx_cut_bool = minx < (vox_size_diag/2)
        gy_cut_bool = miny < (vox_size_diag/2)

        # need to pair the vertices of one group to those of the other
        # for this we actually only need the minima positions of one vertices group
        ### pair the mins of the shorter array onto the longer array
        if len(v_gy) >= len(v_gx):
            vgy_paired = v_gy[minx_pos]
            # limit (cut down) the vertices groups to only those
            # below the defined distance threshold
            vgx_cut = v_gx[gx_cut_bool]
            vgy_cut = vgy_paired[gx_cut_bool]

            V1 = vgx_cut
            V2 = vgy_cut
            V_mean = (V1 + V2) / 2
        else:
            vgx_paired = v_gx[miny_pos]
            vgy_cut = v_gy[gy_cut_bool]
            vgx_cut = vgx_paired[gy_cut_bool]

            V1 = vgx_cut
            V2 = vgy_cut
            V_mean = (V1 + V2) / 2

        return V_mean
    else:
        raise AttributeError(str(ref) + "must be 'x', 'y' or 'mean'")

def get_gradmin_intersect(geo_data, surface_vertices, grad_minima):
    """
        Compute the intersection between gradient minima and a surface of
        interest as vertices.

        Args:
            geo_data (:class:`gempy.data_management.InputData`)
            surface_vertices (ndarray): Vertices of the surface of interest.
            grad_minima (ndarray): Vertices of the gradient minima surface.

        Returns:
            intersection vertices (ndarray)


        """
    vox_size_x = np.abs(geo_data.extent[1] - geo_data.extent[0]) / geo_data.resolution[0]
    vox_size_y = np.abs(geo_data.extent[3] - geo_data.extent[2]) / geo_data.resolution[1]
    vox_size_z = np.abs(geo_data.extent[5] - geo_data.extent[4]) / geo_data.resolution[2]
    vox_size_diag = np.sqrt(vox_size_x ** 2 + vox_size_y ** 2 + vox_size_z ** 2)

    #v_l = np.array(surface_vertices[0])
    v_l = np.array(surface_vertices)
    l_dist = distance.cdist(grad_minima, v_l)
    min_dist = np.min(l_dist, axis=0)
    l_cut_bool = min_dist < (vox_size_diag/2)
    intersect = v_l[l_cut_bool]
    return intersect

def get_voxel_extrema(GX, GY, GZ=np.nan, direction='z'):
    """
        Detects gradient extrema in voxels and classifies them as
        minima, maxima or saddle point based on changes in sign of the
        gradient.

        Args:
            GX (ndarray): Gradient field in x-direction.
            GY (ndarray): Gradient field in y-direction.
            GZ (ndarray, optional): Gradient field in z-direction. Only needed if
                minima are relative to a different direction than 'z'.
            direction (str, optional, default = 'z'): Direction to which the gradients refer to,
                and to which the minima are related.
                Default for standard topography is 'z'.

        Returns:
            All extrema are returned as 'True' or '1' in
            boolean arrays accordingly.

            [0]: minima voxels (ndarray)
            [1]: maxima voxels (ndarray)
            [2]: saddle voxels (ndarray)
        """
    if direction == 'z':
        gx = GX
        gy = GY
    elif direction == 'y':
        gx = GY
        gy = GZ
    elif direction == 'x':
        gx = GZ
        gy = GY
    else:
        raise AttributeError(str(direction) + "must be a cartesian direction, i.e. xyz")

    # getting array with gradient signs (-1 for negative, 1 for positive and exactly 0)
    gx_signs = np.sign(gx)
    gy_signs = np.sign(gy)
    # empty holder arrays for inserting minima and maxima
    gx_maxima = np.zeros_like(gx)
    gy_maxima = np.zeros_like(gy)
    gx_minima = np.zeros_like(gx)
    gy_minima = np.zeros_like(gy)

    # create boolean arrays where signchange voxels are True
    signchange_gx = ((np.roll(gx_signs, 1, axis=0) - gx_signs) != 0).astype(int)
    # avoid border error from np.roll by setting relevant border to False
    signchange_gx[0, :, :] = 0

    signchange_gy = ((np.roll(gy_signs, 1, axis=1) - gy_signs) != 0).astype(int)
    signchange_gy[:, 0] = 0

    # conditions for a voxel to be recognized as maximum or minimum in given direction
    gx_max_cond = (signchange_gx == 1) & (gx_signs == 1)
    gy_max_cond = (signchange_gy == 1) & (gy_signs == 1)
    gx_min_cond = (signchange_gx == 1) & (gx_signs == -1)
    gy_min_cond = (signchange_gy == 1) & (gy_signs == -1)
    # voxels which meet the conditions are marked as True
    # for maxima and minima accordingly
    # since signchange only identifies the next index AFTER the signchange,
    # we include the voxel BEFORE the signchange by doing an according roll
    gx_maxima[gx_max_cond] = 1
    gx_maxima2 = np.roll(gx_maxima, -1, axis=0)
    gx_max_final = gx_maxima + gx_maxima2
    gy_maxima[gy_max_cond] = 1
    gy_maxima2 = np.roll(gy_maxima, -1, axis=1)
    gy_max_final = gy_maxima + gy_maxima2
    # overall maxima in BOTH directions:
    vox_maxima = np.logical_and(gx_max_final, gy_max_final)
    # NOTE: 0 gives a POSITIVE sign! Might have to correct for this?
    # analogous process for minima:
    gx_minima[gx_min_cond] = 1
    gx_minima2 = np.roll(gx_minima, -1, axis=0)
    gx_min_final = gx_minima + gx_minima2
    gy_minima[gy_min_cond] = 1
    gy_minima2 = np.roll(gy_minima, -1, axis=1)
    gy_min_final = gy_minima + gy_minima2
    vox_minima = np.logical_and(gx_min_final, gy_min_final)
    # saddle points as max in one and min in the other direction
    vox_saddles = np.logical_or(np.logical_and(gx_min_final, gy_max_final),
                                np.logical_and(gy_min_final, gx_max_final))

    return vox_minima, vox_maxima, vox_saddles

def get_surface_extrema(geo_data, surface_vertices, GX, GY, ref='x'):
    """
        Compute possible extrema (minima, maxima and saddle points)
        for a surface of interest.
        This works for standard non-extreme topographies,
        but does not consider the presence of df.

        Note: This will currently return various points and also
        several points for one extrema. To attain single points,
        further distinction and selection is required.

        Args:
            geo_data (:class:`gempy.data_management.InputData`)
            GX (ndarray): Gradient field in x-direction.
            GY (ndarray): Gradient field in y-direction.
            surface_vertices (ndarray): Vertices of the surface of interest.
            ref (str, optional, default='x'): Reference direction for
                marching cubes to attain vertices. Default is 'x'.
                'x' or 'y' should suffice and lead to the same results
                in most cases.
                If needed, ref can be defined as the 'mean' of both,
                but this comes at high computational costs and is
                not recommended in most cases.

        Returns:
            3D coordinates for possible extrema.

            [0]: minima coordinates (ndarray)
            [1]: maxima coordinates (ndarray)
            [2]: saddle coordinates (ndarray)
        """
    vox_size_x = np.abs(geo_data.extent[1] - geo_data.extent[0]) / geo_data.resolution[0]
    vox_size_y = np.abs(geo_data.extent[3] - geo_data.extent[2]) / geo_data.resolution[1]
    vox_size_z = np.abs(geo_data.extent[5] - geo_data.extent[4]) / geo_data.resolution[2]
    vox_size_diag = np.sqrt(vox_size_x ** 2 + vox_size_y ** 2 + vox_size_z ** 2)

    grad_minima = get_gradient_minima(geo_data, GX,GY, ref)
    intersect = get_gradmin_intersect(geo_data, surface_vertices, grad_minima)
    vox_minima, vox_maxima, vox_saddles = get_voxel_extrema(GX, GY)

    if np.any(vox_minima):
        # get the coordinates for minima, maxima and saddles
        MIN_coord0 = np.argwhere(vox_minima == True)
        # rescale the coordinates to actual size of voxels
        # to use combined with the intersection coordinates from above
        MIN_coord = MIN_coord0
        MIN_coord[:, 0] = (MIN_coord0[:, 0] * vox_size_x) + geo_data.extent[0] # + vox_size_x/2
        MIN_coord[:, 1] = (MIN_coord0[:, 1] * vox_size_y) + geo_data.extent[2] # + vox_size_y/2
        MIN_coord[:, 2] = (MIN_coord0[:, 2] * vox_size_z) + geo_data.extent[4] # + vox_size_z/2
        # get distances between intersection and the according extrema coordinates
        dist_MIN = distance.cdist(intersect, MIN_coord, 'euclidean')
        # classify intersection extrema by limiting to distance to according voxel coordinates
        # half a voxel-diagonal to get what is "inside" a voxel (best results, yet)
        min_dist_MIN = np.min(dist_MIN, axis=1)
        cut_bool_MIN = min_dist_MIN < (vox_size_diag / 2)
        intersect_minima_all = intersect[cut_bool_MIN]
    else:
        print('No minima found for surface.')
        intersect_minima_all = []

    if np.any(vox_maxima):
        MAX_coord0 = np.argwhere(vox_maxima == True)
        MAX_coord = MAX_coord0
        MAX_coord[:, 0] = (MAX_coord0[:, 0] * vox_size_x) + geo_data.extent[0] # + vox_size_x/2
        MAX_coord[:, 1] = (MAX_coord0[:, 1] * vox_size_y) + geo_data.extent[2] # + vox_size_y/2
        MAX_coord[:, 2] = (MAX_coord0[:, 2] * vox_size_z) + geo_data.extent[4] # + vox_size_z/2
        dist_MAX = distance.cdist(intersect, MAX_coord, 'euclidean')
        min_dist_MAX = np.min(dist_MAX, axis=1)
        cut_bool_MAX = min_dist_MAX < (vox_size_diag / 2)
        intersect_maxima_all = intersect[cut_bool_MAX]
    else:
        print('No maxima found for surface.')
        intersect_maxima_all = []

    if np.any(vox_saddles):
        SADD_coord0 = np.argwhere(vox_saddles == True)
        SADD_coord = SADD_coord0
        SADD_coord[:, 0] = (SADD_coord0[:, 0] * vox_size_x) + geo_data.extent[0] # + vox_size_x/2
        SADD_coord[:, 1] = (SADD_coord0[:, 1] * vox_size_y) + geo_data.extent[2] # + vox_size_y/2
        SADD_coord[:, 2] = (SADD_coord0[:, 2] * vox_size_z) + geo_data.extent[4] # + vox_size_z/2
        dist_SADD = distance.cdist(intersect, SADD_coord, 'euclidean')
        min_dist_SADD = np.min(dist_SADD, axis=1)
        cut_bool_SADD = min_dist_SADD < (vox_size_diag / 2)
        intersect_saddles_all = intersect[cut_bool_SADD]
    else:
        print('No saddle points found for surface.')
        intersect_saddles_all = []

    return intersect_minima_all, intersect_maxima_all, intersect_saddles_all

def get_highest_saddle_point(geo_data, surface_vertices, GX, GY, ref='x'):
    """
        Compute the maximum saddle point for a surface of interest.
        Useful for simple models with only on main structure.

        Args:
            geo_data (:class:`gempy.data_management.InputData`)
            surface_vertices (ndarray): Vertices of the surface of interest.
            GX (ndarray): Gradient field in x-direction.
            GY (ndarray): Gradient field in y-direction.
            ref (str, optional, default='x'): Reference direction for
                marching cubes to attain vertices. Default is 'x'.
                'x' or 'y' should suffice and lead to the same results
                in most cases.
                If needed, ref can be defined as the 'mean' of both,
                but this comes at high computational costs and is
                not recommended in most cases.

        Returns:
            3D coordinates for the highest saddle point (ndarray).
        """
    intersect_saddles_all = get_surface_extrema(geo_data, surface_vertices, GX, GY, ref)[2]
    if len(intersect_saddles_all) > 0:
        max_SADD = intersect_saddles_all[np.argmax(intersect_saddles_all[:, 2])]
    else:
        max_SADD = []
    return max_SADD

def get_highest_max(geo_data, surface_vertices, GX, GY, ref='x'):
    """
        Compute the highest maximum for a surface of interest.
        Useful for simple models with only on main structure.

        Args:
            geo_data (:class:`gempy.data_management.InputData`)
            surface_vertices (ndarray): Vertices of the surface of interest.
            GX (ndarray): Gradient field in x-direction.
            GY (ndarray): Gradient field in y-direction.
            ref (str, optional, default='x'): Reference direction for
                marching cubes to attain vertices. Default is 'x'.
                'x' or 'y' should suffice and lead to the same results
                in most cases.
                If needed, ref can be defined as the 'mean' of both,
                but this comes at high computational costs and is
                not recommended in most cases.

        Returns:
            3D coordinates for the highest maximum (ndarray).
        """
    intersect_maxima_all = get_surface_extrema(geo_data, surface_vertices, GX, GY, ref)[1]
    if len(intersect_maxima_all) > 0:
        it_final_MAX = intersect_maxima_all[np.argmax(intersect_maxima_all[:, 2])]
    else:
        it_final_MAX = []
    return it_final_MAX

def plot_surface_extrema(geo_data, surface_vertices, GX, GY, ref='x'):
    """
        Plot the surface of interest and detected extrema (blue= minima,
        red = maxima, violet = saddle points).

        Args:
            geo_data (:class:`gempy.data_management.InputData`)
            surface_vertices (ndarray): Vertices of the surface of interest.
            GX (ndarray): Gradient field in x-direction.
            GY (ndarray): Gradient field in y-direction.
            ref (str, optional, default='x'): Reference direction for
                marching cubes to attain vertices. Default is 'x'.
                'x' or 'y' should suffice and lead to the same results
                in most cases.
                If needed, ref can be defined as the 'mean' of both,
                but this comes at high computational costs and is
                not recommended in most cases.

        """
    intersect_minima_all, intersect_maxima_all, intersect_saddles_all \
        = get_surface_extrema(geo_data, surface_vertices, GX, GY, ref)
    v_l = surface_vertices
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(v_l[:, 0], v_l[:, 1], v_l[:, 2], color='k', alpha=0.2)
    #ax.scatter(gradmin[:, 0], gradmin[:, 1], gradmin[:, 2], color='k', alpha=0.2)
    #ax.scatter(intersect[:, 0], intersect[:, 1], intersect[:, 2], color='b')
    if len(intersect_minima_all) > 0:
        ax.scatter(intersect_minima_all[:, 0], intersect_minima_all[:, 1], intersect_minima_all[:, 2], \
                   color='blue',s=200, marker='o')
    if len(intersect_maxima_all) > 0:
        ax.scatter(intersect_maxima_all[:,0],intersect_maxima_all[:,1],intersect_maxima_all[:,2],\
                   color='r', s=200, marker='o')
    if len(intersect_saddles_all) > 0:
        ax.scatter(intersect_saddles_all[:, 0], intersect_saddles_all[:, 1], intersect_saddles_all[:, 2], \
                   color='violet',s=200, marker='o')

    ax.set_xlim(geo_data.extent[0], geo_data.extent[1])
    ax.set_ylim(geo_data.extent[2], geo_data.extent[3])
    ax.set_zlim(geo_data.extent[4], geo_data.extent[5])
    ax.set_xlabel('X [m]')
    ax.set_ylabel('Y [m]')
    ax.set_zlabel('Depth [m]')
    plt.show()


"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.

    Module with classes and methods to perform implicit regional modelling based on
    the potential field method.
    Tested on Ubuntu 16

    Created on 10/11/2019

    @author: Alex Schaaf, Elisa Heim, Miguel de la Varga
"""

# This is for sphenix to find the packages
# sys.path.append( path.dirname( path.dirname( path.abspath(__file__) ) ) )

from typing import Union, List, Any

import matplotlib.pyplot as plt
# from .vista import Vista
# import gempy as _gempy
import numpy as np
import pandas as pn
from gempy_legacy.plot.vista import GemPyToVista

# Keep Alex code hidden until we merge it properly
try:
    import pyvista as pv
    from ._vista import Vista as Vista

    PYVISTA_IMPORT = True
except ImportError:
    PYVISTA_IMPORT = False

from .visualization_2d import Plot2D

try:
    import mplstereonet

    mplstereonet_import = True
except ImportError:
    mplstereonet_import = False


def plot_2d(model, n_axis=None, section_names: list = None,
            cell_number: int = None, direction: str = 'y',
            show_data: Union[bool, list] = True,
            show_results: Union[bool, list] = True,
            show_lith: Union[bool, list] = True,
            show_values: Union[bool, list] = False,
            show_block: Union[bool, list] = False,
            show_scalar: Union[bool, list] = False,
            show_boundaries: Union[bool, list] = True,
            show_topography: Union[bool, list] = False,
            show_section_traces: Union[bool, list] = True,
            series_n: Union[int, List[int]] = 0,
            ve=1,
            block=None,
            regular_grid=None,
            kwargs_topography=None,
            kwargs_regular_grid=None,
            **kwargs):
    """Plot 2-D sections of geomodel.

    Plot cross sections either based on custom section traces or cell number in xyz direction.
    Options to plot lithology block, scalar field or rendered surface lines.
    Input data and topography can be included.

    Args:
        show_block (bool): If True and model has been computed, plot cross section
         of the final model.
        show_values (bool): If True and model has been computed, plot cross section
         of the value... TODO need to add attribute to choose which value to be plot
        model: Geomodel object with solutions.
        n_axis (int): Subplot axis for multiple sections
        section_names (list): Names of predefined custom section traces
        cell_number (list): Position of the array to plot
        direction (str): Cartesian direction to be plotted (xyz)
        show_data (bool): Show original input data. Defaults to True.
        show_results (bool): If False, override show lith, show_calar, show_values
        show_lith (bool): Show lithological block volumes. Defaults to True.
        show_scalar (bool): Show scalar field isolines. Defaults to False.
        show_boundaries (bool): Show surface boundaries as lines. Defaults to True.
        show_topography (bool): Show topography on plot. Defaults to False.
        series_n (int): number of the scalar field.
        ve (float): vertical exageration
        regular_grid (numpy.ndarray): Numpy array of the size of model.grid.regular_grid
        kwargs_topography (dict):
            * fill_contour
            * hillshade (bool): Calculate and add hillshading using elevation data
            * azdeg (float): azimuth of sun for hillshade
            * altdeg (float): altitude in degrees of sun for hillshade


    Keyword Args:
        legend (bool): If True plot legend. Default True
        show (bool): Call matplotlib show

    Returns:
        :class:`gempy.plot.visualization_2d.Plot2D`: Plot2D object

    """
    if kwargs_regular_grid is None:
        kwargs_regular_grid = dict()
    if kwargs_topography is None:
        kwargs_topography = dict()
    if section_names is None and cell_number is None and direction is not None:
        cell_number = ['mid']

    show = kwargs.get('show', True)

    if block is not None:
        import warnings
        regular_grid = block
        warnings.warn('block is going to be deprecated. Use regular grid instead',
                      DeprecationWarning)

    section_names = [] if section_names is None else section_names
    section_names = np.atleast_1d(section_names)
    if cell_number is None:
        cell_number = []
    elif cell_number == 'mid':
        cell_number = ['mid']
    direction = [] if direction is None else direction

    if type(cell_number) != list:
        cell_number = [cell_number]

    if type(direction) != list:
        direction = [direction]

    if n_axis is None:
        n_axis = len(section_names) + len(cell_number)

    if show_results is False:
        show_lith = False
        show_values = False
        show_block = False
        show_scalar = False
        show_boundaries = False

    if type(show_data) is bool:
        show_data = [show_data] * n_axis
    if type(show_lith) is bool:
        show_lith = [show_lith] * n_axis
    if type(show_values) is bool:
        show_values = [show_values] * n_axis
    if type(show_block) is bool:
        show_block = [show_block] * n_axis
    if type(show_scalar) is bool:
        show_scalar = [show_scalar] * n_axis
    if type(show_boundaries) is bool:
        show_boundaries = [show_boundaries] * n_axis
    if type(show_topography) is bool:
        show_topography = [show_topography] * n_axis
    if type(series_n) is int:
        series_n = [series_n] * n_axis

    # init e
    e = 0
    # is 10 and 10 because in the ax pos is the second digit
    n_columns_ = 1 if len(section_names) + len(cell_number) < 2 else 2
    n_columns = n_columns_ * 10  # This is for the axis location syntax
    n_rows = (len(section_names) + len(cell_number)) / n_columns_

    n_columns_ = np.max([n_columns_, 1])
    n_rows = np.max([n_rows, 1])

    p = Plot2D(model, **kwargs)
    p.create_figure(cols=n_columns_, rows=n_rows, **kwargs)

    for e, sn in enumerate(section_names):
        # Check if a plot that fills all pixels is plotted
        _is_filled = False
        assert e < 10, 'Reached maximum of axes'

        ax_pos = (round(n_axis / 2 + 0.1)) * 100 + n_columns + e + 1
        temp_ax = p.add_section(section_name=sn, ax_pos=ax_pos, ve=ve, **kwargs)
        if show_data[e] is True:
            p.plot_data(temp_ax, section_name=sn, **kwargs)
        if show_lith[e] is True and model.solutions.lith_block.shape[0] != 0:
            _is_filled = True
            p.plot_lith(temp_ax, section_name=sn, **kwargs)
        elif show_values[e] is True and model.solutions.values_matrix.shape[0] != 0:
            _is_filled = True
            p.plot_values(temp_ax, series_n=series_n[e], section_name=sn, **kwargs)
        elif show_block[e] is True and model.solutions.block_matrix.shape[0] != 0:
            _is_filled = True
            p.plot_block(temp_ax, series_n=series_n[e], section_name=sn, **kwargs)
        if show_scalar[e] is True and model.solutions.scalar_field_matrix.shape[0] != 0:
            _is_filled = True
            p.plot_scalar_field(temp_ax, series_n=series_n[e], section_name=sn, **kwargs)
        if show_boundaries[e] is True and model.solutions.scalar_field_matrix.shape[0] != 0:
            p.plot_contacts(temp_ax, section_name=sn, **kwargs)
        if show_topography[e] is True:
            # Check if anything dense is plot. If not plot dense topography
            f_c_ = not _is_filled
            # f_c = kwargs_topography.get('fill_contour', f_c_)
            if 'fill_contour' not in kwargs_topography:
                kwargs_topography['fill_contour'] = f_c_
            p.plot_topography(temp_ax, section_name=sn,  # fill_contour=f_c,
                              **kwargs_topography)
            if show_section_traces is True and sn == 'topography':
                p.plot_section_traces(temp_ax)

        if regular_grid is not None:
            p.plot_regular_grid(temp_ax, block=regular_grid, section_name=sn,
                                **kwargs_regular_grid)

        temp_ax.set_aspect(ve)

        # If there are section we need to shift one axis for the perpendicular
        e = e + 1

    for e2 in range(len(cell_number)):
        assert (e + e2) < 10, 'Reached maximum of axes'

        ax_pos = (round(n_axis / 2 + 0.1)) * 100 + n_columns + e + e2 + 1
        temp_ax = p.add_section(cell_number=cell_number[e2],
                                direction=direction[e2], ax_pos=ax_pos, ve=ve)
        if show_data[e + e2] is True:
            p.plot_data(temp_ax, cell_number=cell_number[e2],
                        direction=direction[e2], **kwargs)
        if show_lith[e + e2] is True and model.solutions.lith_block.shape[0] != 0:
            p.plot_lith(temp_ax, cell_number=cell_number[e2],
                        direction=direction[e2], **kwargs)
        elif show_values[e + e2] is True and model.solutions.values_matrix.shape[0] != 0:
            p.plot_values(temp_ax, series_n=series_n[e], cell_number=cell_number[e2],
                          direction=direction[e2], **kwargs)
        elif show_block[e + e2] is True and model.solutions.block_matrix.shape[0] != 0:
            p.plot_block(temp_ax, series_n=series_n[e], cell_number=cell_number[e2],
                         direction=direction[e2], **kwargs)
        if show_scalar[e + e2] is True and model.solutions.scalar_field_matrix.shape[0] != 0:
            p.plot_scalar_field(temp_ax, series_n=series_n[e], cell_number=cell_number[e2],
                                direction=direction[e2], **kwargs)
        if show_boundaries[e + e2] is True and model.solutions.scalar_field_matrix.shape[0] != 0:
            p.plot_contacts(temp_ax, cell_number=cell_number[e2],
                            direction=direction[e2], **kwargs)
        if show_topography[e + e2] is True:
            p.plot_topography(temp_ax, cell_number=cell_number[e2],
                              direction=direction[e2], **kwargs_topography)
        if regular_grid is not None:
            p.plot_regular_grid(temp_ax, block=regular_grid, cell_number=cell_number[e2],
                                direction=direction[e2], **kwargs_regular_grid)

        temp_ax.set_aspect(ve)

    if show is True:
        p.fig.show()

    return p


def plot_3d(model, plotter_type='basic',
            show_data: bool = True,
            show_results: bool = True,
            show_surfaces: bool = True,
            show_lith: bool = True,
            show_scalar: bool = False,
            show_boundaries: bool = True,
            show_topography: Union[bool, list] = False,
            scalar_field: str = None,
            ve=None,
            kwargs_plot_structured_grid=None,
            kwargs_plot_topography=None,
            kwargs_plot_data=None,
            image=False,
            off_screen=False,
            font_size: int = pv.global_theme.font.size,
            **kwargs) -> GemPyToVista:
    """foobar

    Args:

        model (:class:`gempy.core.model.Project`): Container class of all
         objects that constitute a GemPy model.
        plotter_type: PyVista plotter types. Supported plotters are:
         'basic', 'background', and 'notebook'.
        show_data (bool): Show original input data. Defaults to True.
        show_results (bool): If False, override show lith, show_scalar, show_values
        show_lith (bool): Show lithological block volumes. Defaults to True.
        show_scalar (bool): Show scalar field isolines. Defaults to False.
        show_boundaries (bool): Show surface boundaries as lines. Defaults to True.
        show_topography (bool): Show topography on plot. Defaults to False.
        scalar_field (str): Name of the field to be activated
        series_n (int): number of the scalar field.
        ve (float): Vertical Exaggeration
        kwargs_plot_structured_grid:
        kwargs_plot_topography:
        font_size : int
            Font size for the labels of the grid
        **kwargs:

    Returns:
        :class:`gempy.plot.vista.GemPyToVista`

    """
    if image is True:
        off_screen = True
        kwargs['off_screen'] = True
        plotter_type = 'basic'
    if show_results is False:
        show_surfaces = False
        show_scalar = False
        show_lith = False

    if kwargs_plot_topography is None:
        kwargs_plot_topography = dict()
    if kwargs_plot_structured_grid is None:
        kwargs_plot_structured_grid = dict()
    if kwargs_plot_data is None:
        kwargs_plot_data = dict()

    fig_path: str = kwargs.get('fig_path', None)

    gpv = GemPyToVista(model, plotter_type=plotter_type, **kwargs)
    if show_surfaces and len(model.solutions.vertices) != 0:
        gpv.plot_surfaces()
    if show_lith is True and model.solutions.lith_block.shape[0] != 0:
        gpv.plot_structured_grid('lith', **kwargs_plot_structured_grid)
    if show_scalar is True and model.solutions.scalar_field_matrix.shape[0] != 0:
        gpv.plot_structured_grid("scalar", series=scalar_field)

    if show_data:
        gpv.plot_data(**kwargs_plot_data)

    if show_topography and model._grid.topography is not None:
        gpv.plot_topography(**kwargs_plot_topography)

    if ve is not None:
        gpv.p.set_scale(zscale=ve)

    if font_size is not None:
        gpv.p.show_bounds(font_size=font_size)

    if fig_path is not None:
        gpv.p.show(screenshot=fig_path)

    if image is True:
        img = gpv.p.show(screenshot=True)
        img = gpv.p.last_image
        plt.imshow(img[1])
        plt.axis('off')
        plt.show(block=False)
        gpv.p.close()

    if off_screen is False:
        gpv.p.show()

    return gpv


def plot_interactive_3d(
        geo_model,
        scalar_field: str = 'all',
        series=None,
        show_topography: bool = False,
        **kwargs,
):
    """Plot interactive 3-D geomodel with three cross sections in subplots.
    Args:
        geo_model: Geomodel object with solutions.
        name (str): Can be either one of the following
                'lith' - Lithology id block.
                'scalar' - Scalar field block.
                'values' - Values matrix block.
        render_topography: Render topography. Defaults to False.
        **kwargs:
    Returns:
        :class:`gempy.plot.vista.GemPyToVista`

    """
    gpv = GemPyToVista(geo_model, plotter_type='background', shape="1|3")
    gpv.plot_data()
    gpv.plot_structured_grid_interactive(scalar_field=scalar_field, series=series,
                                         render_topography=show_topography, **kwargs)

    return gpv


def plot_section_traces(model):
    """Plot section traces of section grid in 2-D topview (xy).

    Args:
        model: Geomodel object with solutions.

    Returns:
        (Plot2D) Plot2D object
    """
    pst = plot_2d(model, n_axis=1, direction=['z'], cell_number=[-1],
                  show_data=False, show_boundaries=False, show_lith=False, show=False)
    pst.plot_section_traces(pst.axes[0], show_data=False)
    return pst


def plot_stereonet(self, litho=None, planes=True, poles=True,
                   single_plots=False,
                   show_density=False):
    if mplstereonet_import is False:
        raise ImportError(
            'mplstereonet package is not installed. No stereographic projection available.')

    from collections import OrderedDict

    if litho is None:
        litho = self.model._orientations.df['surface'].unique()

    if single_plots is False:
        fig, ax = mplstereonet.subplots(figsize=(5, 5))
        df_sub2 = pn.DataFrame()
        for i in litho:
            df_sub2 = df_sub2.append(self.model._orientations.df[
                                         self.model._orientations.df[
                                             'surface'] == i])

    for formation in litho:
        if single_plots:
            fig = plt.figure(figsize=(5, 5))
            ax = fig.add_subplot(111, projection='stereonet')
            ax.set_title(formation, y=1.1)

        # if series_only:
        # df_sub = self.model.orientations.df[self.model.orientations.df['series'] == formation]
        # else:
        df_sub = self.model._orientations.df[
            self.model._orientations.df['surface'] == formation]

        if poles:
            ax.pole(df_sub['azimuth'] - 90, df_sub['dip'], marker='o',
                    markersize=7,
                    markerfacecolor=self._color_lot[formation],
                    markeredgewidth=1.1, markeredgecolor='gray',
                    label=formation + ': ' + 'pole point')
        if planes:
            ax.plane(df_sub['azimuth'] - 90, df_sub['dip'],
                     color=self._color_lot[formation],
                     linewidth=1.5, label=formation + ': ' + 'azimuth/dip')
        if show_density:
            if single_plots:
                ax.density_contourf(df_sub['azimuth'] - 90, df_sub['dip'],
                                    measurement='poles', cmap='viridis',
                                    alpha=.5)
            else:
                ax.density_contourf(df_sub2['azimuth'] - 90, df_sub2['dip'],
                                    measurement='poles', cmap='viridis',
                                    alpha=.5)

        fig.subplots_adjust(top=0.8)
        handles, labels = ax.get_legend_handles_labels()
        by_label = OrderedDict(zip(labels, handles))
        ax.legend(by_label.values(), by_label.keys(), bbox_to_anchor=(1.9, 1.1))
        ax._grid(True, color='black', alpha=0.25)


def plot_topology(geo_model, edges, centroids, direction="y", scale=True,
                  label_kwargs=None, edge_kwargs=None):
    """Plot the topology adjacency graph in 2-D.

        Args:
            geo_model ([type]): GemPy geomodel instance.
            edges (Set[Tuple[int, int]]): Set of topology edges.
            centroids (Dict[int, Array[int, 3]]): Dictionary of topology id's and
                their centroids.
            direction (Union["x", "y", "z", optional): Section direction.
                Defaults to "y".
            label_kwargs (dict, optional): Keyword arguments for topology labels.
                Defaults to None.
            edge_kwargs (dict, optional): Keyword arguments for topology edges.
                Defaults to None.

        """
    res = geo_model._grid.regular_grid.resolution
    if direction == "y":
        c1, c2 = (0, 2)
        e1 = geo_model._grid.regular_grid.extent[1] - geo_model._grid.regular_grid.extent[0]
        e2 = geo_model._grid.regular_grid.extent[5] - geo_model._grid.regular_grid.extent[4]
        d1 = geo_model._grid.regular_grid.extent[0]
        d2 = geo_model._grid.regular_grid.extent[4]
        # if len(list(centroids.items())[0][1]) == 2:
        #     c1, c2 = (0, 1)
        r1 = res[0]
        r2 = res[2]
    elif direction == "x":
        c1, c2 = (1, 2)
        e1 = geo_model._grid.regular_grid.extent[3] - geo_model._grid.regular_grid.extent[2]
        e2 = geo_model._grid.regular_grid.extent[5] - geo_model._grid.regular_grid.extent[4]
        d1 = geo_model._grid.regular_grid.extent[2]
        d2 = geo_model._grid.regular_grid.extent[4]
        # if len(list(centroids.items())[0][1]) == 2:
        #     c1, c2 = (0, 1)
        r1 = res[1]
        r2 = res[2]
    elif direction == "z":
        c1, c2 = (0, 1)
        e1 = geo_model._grid.regular_grid.extent[1] - geo_model._grid.regular_grid.extent[0]
        e2 = geo_model._grid.regular_grid.extent[3] - geo_model._grid.regular_grid.extent[2]
        d1 = geo_model._grid.regular_grid.extent[0]
        d2 = geo_model._grid.regular_grid.extent[2]
        # if len(list(centroids.items())[0][1]) == 2:
        #     c1, c2 = (0, 1)
        r1 = res[0]
        r2 = res[1]

    tkw = {
        "color": "white",
        "fontsize": 13,
        "ha": "center",
        "va": "center",
        "weight": "ultralight",
        "family": "monospace",
        "verticalalignment": "center",
        "horizontalalignment": "center",
        "bbox": dict(boxstyle='round', facecolor='black', alpha=1),
    }
    if label_kwargs is not None:
        tkw.update(label_kwargs)

    lkw = {
        "linewidth": 1,
        "color": "black"
    }
    if edge_kwargs is not None:
        lkw.update(edge_kwargs)

    for a, b in edges:
        # plot edges
        x = np.array([centroids[a][c1], centroids[b][c1]])
        y = np.array([centroids[a][c2], centroids[b][c2]])
        if scale:
            x = x * e1 / r1 + d1
            y = y * e2 / r2 + d2
        plt.plot(x, y, **lkw)

    for node in np.unique(list(edges)):
        x = centroids[node][c1]
        y = centroids[node][c2]
        if scale:
            x = x * e1 / r1 + d1
            y = y * e2 / r2 + d2
        plt.text(x, y, str(node), **tkw)


def plot_ar(geo_model, path=None, project_name=None, api_token=None, secret=None):
    """ Create, upload and retrieve tag to visualize the model in AR in rexview

    https://www.rexos.org/getting-started/

    Args:
        geo_model (gempy.Model):
        path: Location for rex files. Default cwd
        project_name: Name of the project in rexos
        api_token: rexos api token
        secret: rexos secret

    Returns:
        gempy.addons.rex_api.Rextag
    """
    from gempy_legacy.addons.rex_api import upload_to_rexcloud
    from gempy_legacy.addons.gempy_to_rexfile import write_rex, geomodel_to_rex
    if project_name is None:
        project_name = geo_model.meta.project_name

    if path is None:
        path = './'

    rex_bytes = geomodel_to_rex(geo_model)
    files_path = write_rex(rex_bytes, path)
    project_name_ = project_name
    for i in range(40):
        try:
            tag = upload_to_rexcloud(files_path, project_name=project_name_,
                                     api_token=api_token, secret=secret)
            break
        except ConnectionError:
            project_name_ = project_name + str(i)
            pass

    return tag



# Importing GemPy
import gempy_legacy as gp
from gempy_legacy.assets.geophysics import MagneticsPreprocessing

# Importing auxiliary libraries
import numpy as np
import pytest


def test_magnetics_api(interpolator_magnetics):
    # TODO add the check

    geo_model = gp.create_model('test_center_grid_slicing')
    geo_model.set_default_surfaces()
    geo_model.add_surface_points(X=-1, Y=0, Z=0, surface='surface1')
    geo_model.add_surface_points(X=1, Y=0, Z=0, surface='surface1')
    geo_model.add_orientations(X=0, Y=0, Z=0, surface='surface1', pole_vector=(0, 0, 1))
    geo_model._surfaces.add_surfaces_values([0.037289, 0.0297], ['susceptibility'])

    # needed constants
    mu_0 = 4.0 * np.pi * 10e-7  # magnetic permeability in free space [N/A^2]
    cm = mu_0 / (4.0 * np.pi)  # constant for SI unit
    incl = 77.0653  # NOAA
    decl = 6.8116
    B_ext = 52819.8506939139e-9  # T

    geo_model.set_regular_grid(extent=[-5, 5, -5, 5, -5, 5], resolution=[5, 5, 5])
    geo_model.set_centered_grid(np.array([[0, 0, 0]]), resolution=[10, 10, 15], radius=5000)
    geo_model.set_aesara_function(interpolator_magnetics)
    geo_model._interpolator.set_aesara_shared_magnetics(V='auto', pos_magnetics=1,
                                                        incl= incl, decl=decl, B_ext=B_ext)

    gp.compute_model(geo_model)
    print(geo_model._interpolator.aesara_graph.lg0.get_value())
    print(geo_model.solutions.fw_magnetics)
    np.testing.assert_almost_equal(geo_model.solutions.fw_magnetics,
                                   np.array([473.7836]), decimal=4)


@pytest.fixture(scope="module")
def test_magnetics_no_regular_grid(interpolator_magnetics):
    # TODO add the check

    geo_model = gp.create_model('test_center_grid_slicing')
    geo_model.set_default_surfaces()
    geo_model.add_surface_points(X=-1, Y=0, Z=0, surface='surface1')
    geo_model.add_surface_points(X=1, Y=0, Z=0, surface='surface1')
    geo_model.add_orientations(X=0, Y=0, Z=0, surface='surface1', pole_vector=(0, 0, 1))
    geo_model._surfaces.add_surfaces_values([0.037289, 0.0297], ['susceptibility'])

    # needed constants
    mu_0 = 4.0 * np.pi * 10e-7  # magnetic permeability in free space [N/A^2]
    cm = mu_0 / (4.0 * np.pi)  # constant for SI unit
    incl = 77.0653  # NOAA
    decl = 6.8116
    B_ext = 52819.8506939139e-9  # T

    geo_model.set_centered_grid(np.array([0, 0, 0]), resolution=[10, 10, 15], radius=5000)

    Vmodel = MagneticsPreprocessing(geo_model._grid.centered_grid).set_Vs_kernel()
    # gp.set_interpolator(geo_model, output=['magnetics'])
    geo_model.set_aesara_function(interpolator_magnetics)
    geo_model._interpolator.set_aesara_shared_magnetics(V= Vmodel, pos_magnetics=1,
                                                        incl= incl, decl=decl, B_ext=B_ext)

    # geo_model.interpolator.aesara_graph.V.set_value(Vmodel)
    # geo_model.interpolator.aesara_graph.incl.set_value(incl)
    # geo_model.interpolator.aesara_graph.decl.set_value(decl)
    # geo_model.interpolator.aesara_graph.B_ext.set_value(B_ext)

    gp.compute_model(geo_model)
    np.testing.assert_almost_equal(geo_model.solutions.fw_magnetics,
                                   np.array([473.7836]), decimal=4)
    
    return geo_model


def test_center_grid_slicing(test_magnetics_no_regular_grid):
    geo_model = test_magnetics_no_regular_grid

    geo_model.set_centered_grid(np.array([[0, 0, 0],
                                [1, 1, 1]]), resolution=[10, 10, 15], radius=5000)

    gp.compute_model(geo_model)
    print(geo_model._interpolator.aesara_graph.lg0.get_value())



from setuptools import setup, find_packages, Extension

version = '2.3.2'

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='gempy_legacy',
    version=version,
    packages=find_packages(exclude=('test', 'docs', 'examples')),
    include_package_data=True,
    install_requires=[
        'pandas==2.0.1',
        'aesara',
        'pymc',
        'matplotlib',
        'numpy',
        'pytest',
        'seaborn>=0.9',
        'networkx',
        'scikit-image>=0.17',
        'scikit-learn',
        'pyvista==0.39.1',
        'pyvistaqt==0.10.0',
        'pyqt5',
        'iPython',
        'xarray'
    ],
    url='https://github.com/cgre-aachen/gempy',
    license='LGPL v3',
    author='Miguel de la Varga, Alexander Zimmerman, Elisa Heim, Alexander Schaaf, Fabian Stamm, Florian Wellmann, Jan Niederau, Andrew Annex, Alexander Juestel',
    author_email='miguel@terranigma-solutions.com',
    description='An Open-source, Python-based 3-D structural geological modeling software.',
    long_description_content_type='text/markdown',
    long_description=long_description,
    keywords=['geology', '3-D modeling', 'structural geology', 'uncertainty']
)



import pytest

import gempy_legacy as gp
import pandas as pd
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt


def test_issue_566():
    from pyvista import set_plot_theme
    set_plot_theme('document')

    geo_model = gp.create_model('Model1')
    geo_model = gp.init_data(geo_model, extent=[0, 791, 0, 200, -582, 0],
                             resolution=[100, 10, 100])

    geo_model.set_default_surfaces()
    geo_model.add_surface_points(X=223, Y=0.01, Z=-94, surface='surface1')


@pytest.mark.skip(reason="Expensive test. It should be ran manually.")
def test_issue_569(data_path):
    surface_points_df = df = pd.read_csv(data_path + "/coordinates_mwe.csv")
    orientations_df = pd.read_csv(data_path + "/orientations_mwe.csv")

    geo_model = gp.create_model("Deltatest")
    gp.init_data(geo_model,
                 [df.X.min() - 50, df.X.max() + 50, df.Y.min() - 50, df.Y.max() + 50,
                  df.Z.min() - 50, df.Z.max() + 50, ], [50, 50, 50],
                 surface_points_df=surface_points_df,
                 orientations_df=orientations_df,
                 default_values=True)

    fault_list = []
    series = {"Strat_Series": surface_points_df.loc[
        ~surface_points_df["formation"].str.contains(
            "fault"), "formation"].unique().tolist()}

    for fault in surface_points_df.loc[
        surface_points_df["formation"].str.contains("fault"), "formation"].unique():
        series[fault] = fault
        fault_list.append(fault)

    gp.map_stack_to_surfaces(geo_model,
                             series,
                             remove_unused_series=True)

    geo_model.set_is_fault(fault_list)

    geo_model.reorder_features(['fault_a', 'fault_b', 'Strat_Series'])
    geo_model.add_surfaces("basement")

    plot = gp.plot_2d(geo_model, show_lith=False, show_boundaries=True,
                      direction=['z'])
    plt.show(block=False)

    gp.set_interpolator(geo_model,
                        compile_aesara=True,
                        aesara_optimizer='fast_compile',
                        )
    gp.get_data(geo_model, 'kriging')

    sol = gp.compute_model(geo_model, sort_surfaces=True)

    gp.plot_2d(geo_model, show_scalar=True, series_n=0)
    gp.plot_2d(geo_model, series_n=0)

    gp.plot_3d(geo_model, image=True)


@pytest.mark.skip(reason="Private data missing from the repo")
def test_issue_564(data_path):
    geo_model = gp.create_model('SBPM')

    gp.init_data(geo_model, [550, 730, -200, 1000, 20, 55], [50, 50, 50],
                 path_i=data_path + "/564_Points.csv",
                 path_o=data_path + "/564_Orientations_.csv",
                 default_values=True)

    gp.map_stack_to_surfaces(geo_model,
                             {"Q": 'Quartaer',
                              "vK": 'verwKeuper',
                              "Sa": 'Sandstein',
                              "Sc": 'Schluffstein',
                              "b": 'basement'},
                             remove_unused_series=True)

    gp.set_interpolator(geo_model,
                        compile_aesara=True,
                        aesara_optimizer='fast_compile')

    sol = gp.compute_model(geo_model)
    gp.plot_2d(geo_model)

    gpv = gp.plot_3d(geo_model, image=True, plotter_type='basic', ve=5,
                     show_lith=False)

    geo_model.set_bottom_relation(["Q", "vK", "Sa", "Sc", "b"],
                                  ["Onlap", "Onlap", "Onlap", "Onlap",
                                   "Onlap"])

    sol = gp.compute_model(geo_model)
    gp.plot_2d(geo_model)

    gpv = gp.plot_3d(geo_model, image=True, plotter_type='basic', ve=5,
                     show_lith=True)




from pylab import *
import copy
import pandas as pn
import gempy_legacy as gp
import numpy as np

try:
    import xml.etree.cElementTree as ET
except ImportError:
    import xml.etree.ElementTree as ET


class ReadGeoModellerXML:
    def __init__(self, fp):
        """
        Reads in and parses a GeoModeller XML file to extract interface and orientation data and the overall model
        settings (e.g. extent and sequential pile). It uses ElementTree to parse the XML and the tree's root can
        be accessed using self.root for more direct access to the file.

        Todo: - extract faults

        Args:
            fp (str): Filepath for the GeoModeller xml file to be read.

        """
        self.tree = ET.ElementTree(file=fp)  # load xml as tree
        self.root = self.tree.getroot()

        self.xmlns = "http://www.geomodeller.com/geo"
        self.gml = "http://www.opengis.net/gml"

        self.extent = self._get_extent()
        self.data = self.extract_data()

        self.series = list(self.data.keys())
        self.stratigraphic_column, self.surface_points, self.orientations = self.get_dataframes()

        self.series_info = self._get_series_fmt_dict()

        self.faults = self.get_faults()

        self.series_distribution = self.get_series_distribution()

        self.fault_matrix = self.get_fault_matrix()

    def _get_extent(self):
        """
        Extracts model extent from ElementTree root and returns it as tuple of floats.

        Returns:
            tuple: Model extent as (xmin, xmax, ymin, ymax, zmin, zmax).
        """
        xy = self.root[0][0][0][0].attrib
        z = self.root[0][0][0][1].attrib
        return tuple(np.array([xy["Xmin"], xy["Xmax"],
                               xy["Ymin"], xy["Ymax"],
                               z["Zmin"], z["Zmax"]]).astype(float))

    def extract_data(self):
        """
        Extracts relevant data from the GeoModeller XML file ElementTree root (self.root) and returns it as a dictionary.


        Returns:
            (dict): Data dictionary
        """
        data = {}
        for s in self.get_psc():
            sn = s.get("name")
            data[sn] = {}  # create a dict for each series
            data[sn]["formations"] = []
            data[sn]["InfluencedByFault"] = []
            data[sn]["relation"] = s.get("relation")  # add relation, whatever that is

            for c in s:
                if c.tag == "{" + self.xmlns + "}Data":  # append formation names to list of formations
                    data[sn]["formations"].append(c.get("Name"))

                if c.tag == "{" + self.xmlns + "}InfluencedByFault":  # add fault influences
                    data[sn]["InfluencedByFault"].append(c.get("Name"))

                if c.tag == "{" + self.xmlns + "}PotentialField":

                    data[sn]["gradients"] = []
                    data[sn]["interfaces"] = []
                    data[sn]["interfaces_counters"] = []
                    data[sn]["solutions"] = []
                    data[sn]["constraints"] = []

                    for cc in c:
                        # COVARIANCE
                        if cc.tag == "{" + self.xmlns + "}covariance":
                            data[sn]["covariance"] = cc.attrib

                        # GRADIENTS
                        if cc.tag == "{" + self.xmlns + "}Gradients":
                            for gr in cc:
                                data[sn]["gradients"].append([gr.get("Gx"), gr.get("Gy"), gr.get("Gz"),
                                                              gr.get("XGr"), gr.get("YGr"), gr.get("ZGr")])

                        # INTERFACES
                        if cc.tag == "{" + self.xmlns + "}Points":
                            for co in cc:
                                data[sn]["interfaces"].append([float(co[0].text), float(co[1].text), float(co[2].text)])

                        # INTERFACE COUNTERS
                        if cc.tag == "{" + self.xmlns + "}InterfacePoints":
                            for ip in cc:
                                data[sn]["interfaces_counters"].append([int(ip.get("npnt")), int(ip.get("pnt"))])

                        # CONSTRAINTS
                        if cc.tag == "{" + self.xmlns + "}Constraints":
                            for co in cc:
                                data[sn]["constraints"].append(float(co.get("value")))

                        # SOLUTIONS
                        if cc.tag == "{" + self.xmlns + "}Solutions":
                            for sol in cc:
                                data[sn]["solutions"].append(float(sol.get("sol")))

                        if cc.tag == "{" + self.xmlns + "}ModelFaults":
                            print('hey')

                    # convert from str to float
                    data[sn]["gradients"] = np.array(data[sn]["gradients"]).astype(float)
                    data[sn]["interfaces"] = np.array(data[sn]["interfaces"]).astype(float)
                    data[sn]["interfaces_counters"] = np.array(data[sn]["interfaces_counters"]).astype(float)
                    data[sn]["solutions"] = np.array(data[sn]["solutions"]).astype(float)

        return data

    def get_dataframes(self):

        strat_pile = dict.fromkeys(self.series)
        surface_points = pn.DataFrame()
        orientations = pn.DataFrame()

        for serie in self.series:
            strat_pile[serie] = self.data[serie]['formations']

            interf_s = self.data[serie].get('interfaces')
            orient_s = self.data[serie].get('gradients')

            formations = self.data[serie].get('formations')
            if interf_s is not None:
                interf = pn.DataFrame(columns=['X', 'Y', 'Z'], data=interf_s)
                interf['series'] = serie

                if len(formations) > 1:
                    interf_formations = []
                    for j, fmt in enumerate(formations):
                        for n in range(int(self.data[serie].get('interfaces_counters')[j, 0])):
                            interf_formations.append(fmt)
                    interf['formation'] = interf_formations

                else:
                    interf['formation'] = formations[0]

                surface_points = pn.DataFrame.append(surface_points, interf)

            if orient_s is not None:
                orient = pn.DataFrame(columns=['G_x', 'G_y', 'G_z', 'X', 'Y', 'Z'], data=orient_s)
                orient['series'] = serie
                orient['formation'] = formations[0]  # formation is wrong here but does not matter for orientations

                orientations = pn.DataFrame.append(orientations, orient)

        return strat_pile, surface_points, orientations

    def get_psc(self):
        """Returns the ProjectStratigraphicColumn tree element used for several data extractions."""
        return self.root.find("{" + self.xmlns + "}GeologicalModel").find(
            "{" + self.xmlns + "}ProjectStratigraphicColumn")


    def get_order_formations(self):
        order_formations = []
        for entry in self.series_distribution.values():
            if type(entry) is str:
                order_formations.append(entry)
            elif type(entry) is tuple:
                for e in entry:
                    order_formations.append(e)

        return order_formations

    def get_faults(self):
        """
        Extracts fault names from ElementTree root.

        Returns:
            tuple: Fault names (str) ordered as in the GeoModeller XML.
        """
        faults = []
        for c in self.root[2]:
            faults.append(c.get("Name"))
        return tuple(faults)

    def get_series_distribution(self):
        """
        Combines faults and stratigraphic series into an unordered dictionary as keys and maps the correct
        formations to them as a list value. Faults series get a list of their own string assigned as formation.

        Returns:
            (dict): maps Series (str) -> Formations (list of str)
        """
        series_distribution = {}
        for key in self.series_info.keys():
            fmts = self.series_info[key]["formations"]
            if len(fmts) == 1:
                series_distribution[key] = fmts[0]
            else:
                series_distribution[key] = tuple(fmts)

        for f in self.stratigraphic_column:
            if "Fault" in f or "fault" in f:
                series_distribution[f] = f

        return series_distribution

    def _get_series_fmt_dict(self):
        sp = {}
        for i, s in enumerate(self.stratigraphic_column):  # loop over all series
            fmts = []  # init formation storage list
            influenced_by = []  # init influenced by list
            for c in self.root.find("{" + self.xmlns + "}GeologicalModel").find(
                    "{" + self.xmlns + "}ProjectStratigraphicColumn")[i]:
                if "Data" in c.tag:
                    fmts.append(c.attrib["Name"])
                elif "InfluencedByFault" in c.tag:
                    influenced_by.append(c.attrib["Name"])
            # print(fmts)
            sp[s] = {}
            sp[s]["formations"] = fmts
            sp[s]["InfluencedByFault"] = influenced_by

        return sp


    def _where_do_faults_stop(self):
        fstop = {}
        for i, f in enumerate(self.root[2]):

            stops_on = []
            for c in self.root[2][i][2:]:
                stops_on.append(c.get("Name"))

            fstop[f.get("Name")] = stops_on

        return fstop

    def get_fault_matrix(self):
        nf = len(self.faults)
        fm = np.zeros((nf, nf))  # zero matrix of n_faults²
        fstop = self._where_do_faults_stop()
        for i, f in enumerate(self.faults):
            for fs in fstop[f]:
                j = np.where(np.array(self.faults) == fs)[0][0]
                fm[i, j] = 1

        return fm



import gempy_legacy as gp
import matplotlib.pyplot as plt
import pandas as pn
import numpy as np
import os
import pytest
input_path = os.path.dirname(__file__)+'/../input_data'

# ## Preparing the Python environment
#
# For modeling with GemPy, we first need to import it. We should also import any other packages we want to
# utilize in our Python environment.Typically, we will also require `NumPy` and `Matplotlib` when working
# with GemPy. At this point, we can further customize some settings as desired, e.g. the size of figures or,
# as we do here, the way that `Matplotlib` figures are displayed in our notebook (`%matplotlib inline`).


# These two lines are necessary only if GemPy is not installed
import sys, os
sys.path.append("../..")

import sys, os

sys.path.append("../../../gempy")
import gempy_legacy as gp

import numpy as np
import matplotlib.pyplot as plt
import pandas as pd


def test_pile_geomodel(interpolator):
    ve = 3
    extent = [451e3, 456e3, 6.7820e6, 6.7840e6, -2309 * ve, -1651 * ve]

    geo_model = gp.create_model('Topology-Gullfaks')

    gp.init_data(geo_model, extent, [30, 30, 30],
                 path_o=input_path + "/filtered_orientations.csv",
                 path_i=input_path + "/filtered_surface_points.csv", default_values=True)

    series_distribution = {
        "fault3": "fault3",
        "fault4": "fault4",
        "unconformity": "BCU",
        "sediments": ("tarbert", "ness", "etive"),
    }

    gp.map_stack_to_surfaces(geo_model,
                             series_distribution,
                             remove_unused_series=True)

    geo_model.reorder_features(["unconformity", "fault3", "fault4",
                              "sediments", "Basement"])

    geo_model.set_is_fault(["fault3"])
    geo_model.set_is_fault(["fault4"])

    rel_matrix = np.array([[0, 0, 0, 0, 0],
                           [0, 0, 0, 1, 1],
                           [0, 0, 0, 1, 1],
                           [0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0]])

    geo_model.set_fault_relation(rel_matrix)

    surf_groups = pd.read_csv(input_path + "/filtered_surface_points.csv").group
    geo_model._surface_points.df["group"] = surf_groups
    orient_groups = pd.read_csv(input_path + "/filtered_orientations.csv").group
    geo_model._orientations.df["group"] = orient_groups

    geo_model._surface_points.df.reset_index(inplace=True, drop=True)
    geo_model._orientations.df.reset_index(inplace=True, drop=True)

    geo_model.set_aesara_function(interpolator)
    gp.compute_model(geo_model)

    gp.plot.plot_2d(geo_model, cell_number=25,
                    direction='y', show_data=True)

    plt.savefig(os.path.dirname(__file__) + '/../figs/test_pile_lith_block')

    return geo_model


def test_pile_geomodel_2(interpolator):
    ve = 3
    extent = [451e3, 456e3, 6.7820e6, 6.7840e6, -2309 * ve, -1651 * ve]

    geo_model = gp.create_model('Topology-Gullfaks')

    gp.init_data(geo_model, extent, [30, 30, 30],
                 path_o=input_path + "/filtered_orientations.csv",
                 path_i=input_path + "/filtered_surface_points.csv", default_values=True)

    series_distribution = {
        "fault3": "fault3",
        "fault4": "fault4",
        "unconformity": "BCU",
        "sediments": ("tarbert", "ness", "etive"),
    }

    gp.map_stack_to_surfaces(geo_model,
                             series_distribution,
                             remove_unused_series=True)

    geo_model.reorder_features(["unconformity", "fault3", "fault4",
                              "sediments", "Basement"])

    geo_model.set_is_fault(["fault3"])
    geo_model.set_is_fault(["fault4"])

    rel_matrix = np.array([[0, 0, 0, 0, 0],
                           [0, 0, 0, 1, 1],
                           [0, 0, 0, 1, 1],
                           [0, 0, 0, 0, 0],
                           [0, 0, 0, 0, 0]])

    geo_model.set_fault_relation(rel_matrix)

    surf_groups = pd.read_csv(input_path + "/filtered_surface_points.csv").group
    geo_model._surface_points.df["group"] = surf_groups
    orient_groups = pd.read_csv(input_path + "/filtered_orientations.csv").group
    geo_model._orientations.df["group"] = orient_groups

    geo_model._surface_points.df.reset_index(inplace=True, drop=True)
    geo_model._orientations.df.reset_index(inplace=True, drop=True)

    geo_model.set_aesara_function(interpolator)
    gp.compute_model(geo_model)

    gp.plot.plot_2d(geo_model, cell_number=25,
                         direction='y', show_data=True)

    from gempy_legacy.plot.plot_api import plot_2d

    p = plot_2d(geo_model, cell_number=[25])

    plt.savefig(os.path.dirname(__file__) + '/../figs/test_pile_lith_block')

    return geo_model


def test_reorder_series():

    geo_model = gp.create_model('Geological_Model1')
    geo_model = gp.init_data(geo_model, extent=[0, 4000, 0, 2775, 200, 1200], resolution=[100, 10, 100])
    # Adding a fault
    geo_model.rename_features(['Cycle1'])

    geo_model.add_features(['Fault1'])
    geo_model.set_is_fault(['Fault1'])
    geo_model.reorder_features(['Fault1', 'Cycle1'])
    assert (geo_model._stack.df['BottomRelation'] == ['Fault', 'Erosion']).all()
    assert (geo_model._stack.df.index == ['Fault1', 'Cycle1']).all()
    print(geo_model._stack.df)


def test_complete_model(tmpdir, interpolator):
    # ### Initializing the model:
    compute = True

    geo_model = gp.create_model('Geological_Model1')
    geo_model = gp.init_data(geo_model, extent=[0, 4000, 0, 2775, 200, 1200], resolution=[100, 10, 100])

    if compute is True:
        geo_model.set_aesara_function(interpolator)

    from gempy_legacy.plot import visualization_2d as vv

    # In this case perpendicular to the z axes
    p2d = vv.Plot2D(geo_model)
    p2d.create_figure((15, 8))
    ax = p2d.add_section(direction='z', ax_pos=121)

    ax2 = p2d.add_section(direction='y', ax_pos=122)
    ax2.set_xlim(geo_model._grid.regular_grid.extent[0], geo_model._grid.regular_grid.extent[1])
    ax2.set_ylim(geo_model._grid.regular_grid.extent[4], geo_model._grid.regular_grid.extent[5])

    geo_model.add_surfaces(['D', 'C', 'B', 'A'])

    # surface B
    geo_model.add_surface_points(X=584, Y=285, Z=500, surface='B')
    geo_model.add_surface_points(X=494, Y=696, Z=500, surface='B')
    geo_model.add_surface_points(X=197, Y=1898, Z=500, surface='B')
    geo_model.add_surface_points(X=473, Y=2180, Z=400, surface='B')
    geo_model.add_surface_points(X=435, Y=2453, Z=400, surface='B')
    # surface C
    geo_model.add_surface_points(X=946, Y=188, Z=600, surface='C')
    geo_model.add_surface_points(X=853, Y=661, Z=600, surface='C')
    geo_model.add_surface_points(X=570, Y=1845, Z=600, surface='C')
    geo_model.add_surface_points(X=832, Y=2132, Z=500, surface='C')
    geo_model.add_surface_points(X=767, Y=2495, Z=500, surface='C')
    # Surface D
    geo_model.add_surface_points(X=967, Y=1638, Z=800, surface='D')
    geo_model.add_surface_points(X=1095, Y=996, Z=800, surface='D')

    geo_model.add_orientations(X=832, Y=2132, Z=500, surface='C',
                               orientation=[98, 17.88, 1])

    p2d.plot_data(ax, direction='z')
    p2d.plot_data(ax2, direction='y')

    gp.compute_model(geo_model) if compute else None
    p2d.plot_contacts(ax, direction='z', cell_number=-10)
    p2d.plot_lith(ax, direction='z', cell_number=-10)

    p2d.plot_contacts(ax2, direction='y', cell_number=5)
    p2d.plot_lith(ax2, direction='y', cell_number=5)

    plt.savefig(os.path.dirname(__file__) + '/../figs/test_pile_complete')

    # -----------
    # Adding a fault
    geo_model.rename_features(['Cycle1'])

    geo_model.add_features(['Fault1'])
    geo_model.set_is_fault(['Fault1'])

    geo_model.modify_order_features(1, 'Fault1')
    geo_model.add_surfaces(['F1'])
    gp.map_stack_to_surfaces(geo_model, {'Fault1': 'F1'})

    # Add input data of the fault
    geo_model.add_surface_points(X=1203, Y=138, Z=600, surface='F1')
    geo_model.add_surface_points(X=1250, Y=1653, Z=800, surface='F1')
    # Add orientation
    geo_model.add_orientations(X=1280, Y=2525, Z=500, surface='F1', orientation=[272, 90, -1])

    gp.compute_model(geo_model)

    p2d.remove(ax)
    p2d.plot_data(ax, direction='z', cell_number=-10)
    p2d.plot_contacts(ax, direction='z', cell_number=-10)
    p2d.plot_lith(ax, direction='z', cell_number=-10)

    p2d.remove(ax2)
    # Plot
    p2d.plot_data(ax2, cell_number=5)
    p2d.plot_lith(ax2, cell_number=5)
    p2d.plot_contacts(ax2, cell_number=5)

    # surface B
    geo_model.add_surface_points(X=1447, Y=2554, Z=500, surface='B')
    geo_model.add_surface_points(X=1511, Y=2200, Z=500, surface='B')
    geo_model.add_surface_points(X=1549, Y=629, Z=600, surface='B')
    geo_model.add_surface_points(X=1630, Y=287, Z=600, surface='B')
    # surface C
    geo_model.add_surface_points(X=1891, Y=2063, Z=600, surface='C')
    geo_model.add_surface_points(X=1605, Y=1846, Z=700, surface='C')
    geo_model.add_surface_points(X=1306, Y=1641, Z=800, surface='C')
    geo_model.add_surface_points(X=1476, Y=979, Z=800, surface='C')
    geo_model.add_surface_points(X=1839, Y=962, Z=700, surface='C')
    geo_model.add_surface_points(X=2185, Y=893, Z=600, surface='C')
    geo_model.add_surface_points(X=2245, Y=547, Z=600, surface='C')
    # Surface D
    geo_model.add_surface_points(X=2809, Y=2567, Z=600, surface='D')
    geo_model.add_surface_points(X=2843, Y=2448, Z=600, surface='D')
    geo_model.add_surface_points(X=2873, Y=876, Z=700, surface='D')

    # Compute
    gp.compute_model(geo_model)

    # Plot
    p2d.remove(ax)
    p2d.plot_data(ax, direction='z', cell_number=-10)
    p2d.plot_contacts(ax, direction='z', cell_number=-10)
    p2d.plot_lith(ax, direction='z', cell_number=-10)

    p2d.remove(ax2)
    p2d.plot_lith(ax2, cell_number=5)
    p2d.plot_data(ax2, cell_number=5)

    plt.savefig(os.path.dirname(__file__) + '/../figs/test_pile_complete')

    # ----------------
    # Second cycle
    geo_model.add_features(['Cycle2'])
    geo_model.add_surfaces(['G', 'H'])
    gp.map_stack_to_surfaces(geo_model, {'Cycle2': ['G', 'H']})
    geo_model.reorder_features(['Cycle2', 'Fault1', 'Cycle1'])

    # Surface G
    geo_model.add_surface_points(X=1012, Y=1493, Z=900, surface='G')
    geo_model.add_surface_points(X=1002, Y=1224, Z=900, surface='G')
    geo_model.add_surface_points(X=1996, Y=47, Z=800, surface='G')
    geo_model.add_surface_points(X=300, Y=907, Z=700, surface='G')
    # Surface H
    geo_model.add_surface_points(X=3053, Y=727, Z=800, surface='G')
    # Orientation
    geo_model.add_orientations(X=1996, Y=47, Z=800, surface='G', orientation=[272, 5.54, 1])

    # Compute
    gp.compute_model(geo_model)

    # Plot
    p2d.remove(ax)
    p2d.plot_data(ax, direction='z', cell_number=-10)
    p2d.plot_contacts(ax, direction='z', cell_number=-10)
    p2d.plot_lith(ax, direction='z', cell_number=-10)

    p2d.remove(ax2)
    p2d.plot_lith(ax2, cell_number=5)
    p2d.plot_data(ax2, cell_number=5)
    p2d.plot_contacts(ax2, cell_number=5)

    plt.savefig(os.path.dirname(__file__) + '/../figs/test_pile_complete')


    # ----------------
    # Second Fault
    geo_model.add_features('Fault2')
    geo_model.set_is_fault('Fault2')
    geo_model.add_surfaces('F2')

    geo_model.reorder_features(['Cycle2', 'Fault1', 'Fault2', 'Cycle1'])
    gp.map_stack_to_surfaces(geo_model, {'Fault2': 'F2'})

    geo_model.add_surface_points(X=3232, Y=178, Z=1000, surface='F2')
    geo_model.add_surface_points(X=3132, Y=951, Z=700, surface='F2')
    # geo_model.add_surface_points(X=2962, Y=2184, Z=700, surface='F2')

    geo_model.add_orientations(X=3132, Y=951, Z=700, surface='F2', orientation=[95, 90, 1])

    # Compute
    gp.compute_model(geo_model)

    # Plot
    p2d.remove(ax)
    p2d.plot_data(ax, direction='z', cell_number=5, legend='force')
    p2d.plot_lith(ax, direction='z', cell_number=5)
    p2d.plot_contacts(ax, direction='z', cell_number=5)

    p2d.remove(ax2)
    p2d.plot_lith(ax2, cell_number=5)
    p2d.plot_data(ax2, cell_number=5)
    p2d.plot_contacts(ax2, cell_number=5)

    plt.savefig(os.path.dirname(__file__) + '/../figs/test_pile_complete')

    geo_model.add_surface_points(X=3135, Y=1300, Z=700, surface='D')
    geo_model.add_surface_points(X=3190, Y=969, Z=700, surface='D')

    geo_model.add_surface_points(X=3031, Y=2725, Z=800, surface='G')
    geo_model.add_surface_points(X=3018, Y=1990, Z=800, surface='G')
    geo_model.add_surface_points(X=3194, Y=965, Z=700, surface='G')

    geo_model.add_surface_points(X=3218, Y=1818, Z=890, surface='H')
    geo_model.add_surface_points(X=3934, Y=1207, Z=810, surface='H')

    # Compute
    gp.compute_model(geo_model)

    # Plot
    p2d.remove(ax)
    p2d.plot_data(ax, direction='z', cell_number=5, legend='force')
    p2d.plot_lith(ax, direction='z', cell_number=5)
    p2d.plot_contacts(ax, direction='z',cell_number=5)

    p2d.remove(ax2)
    p2d.plot_lith(ax2, cell_number=5)
    p2d.plot_data(ax2, cell_number=5)
    p2d.plot_contacts(ax2, cell_number=5)

    plt.savefig(os.path.dirname(__file__) + '/../figs/test_pile_complete')






from typing import Union

import numpy as np
import pandas as pn

from ..grid import Grid
from .surface_points import SurfacePoints
from .orientations import Orientations
from ...utils import docstring as ds
from ...utils.meta import _setdoc_pro


@_setdoc_pro([SurfacePoints.__doc__, Orientations.__doc__, Grid.__doc__])
class ScalingSystem(object):
    """
    Auxiliary class to rescale the coordinates between 0 and 1 to increase float stability.

    Attributes:
        df (:class:`pn.DataFrame`): Data frame containing the rescaling factor and centers
        surface_points (:class:`SurfacePoints`): [s0]
        orientations (:class:`Orientations`): [s1]
        grid (:class:`Grid`): [s2]

    Args:
        surface_points (:class:`SurfacePoints`):
        orientations (:class:`Orientations`):
        grid (:class:`Grid`):
        rescaling_factor (float): value which divide all coordinates
        centers (list[float]): New center of the coordinates after shifting
    """

    def __init__(self, surface_points: SurfacePoints, orientations: Orientations, grid: Grid,
                 rescaling_factor: float = None, centers: Union[list, pn.DataFrame] = None):

        self.axial_anisotropy = False
        self.max_coord = np.zeros(3)
        self.min_coord = np.zeros(3)
        self.axial_anisotropy_type = 'data'

        self.surface_points = surface_points
        self.orientations = orientations
        self.grid = grid

        self.df = pn.DataFrame(np.array([rescaling_factor, centers]).reshape(1, -1),
                               index=['values'],
                               columns=['rescaling factor', 'centers'])

        self.rescale_data(rescaling_factor=rescaling_factor, centers=centers)

    def __repr__(self):
        return self.df.T.to_string()

    def _repr_html_(self):
        return self.df.T.to_html()

    def toggle_axial_anisotropy(self, type='data'):
        self.axial_anisotropy_type = type
        self.axial_anisotropy = self.axial_anisotropy ^ True
        self.rescale_data()

    @_setdoc_pro([ds.centers, ds.rescaling_factor])
    def modify_rescaling_parameters(self, attribute, value):
        """
        Modify the parameters used to rescale data

        Args:
            attribute (str): Attribute to be modified. It can be: centers, rescaling factor
                * centers: [s0]
                * rescaling factor: [s1]
            value (float, list[float])


        Returns:
            :class:`gempy.core.data_modules.geometric_data.Rescaling`

        """
        assert np.isin(attribute, self.df.columns).all(), 'Valid attributes are: ' + np.array2string(self.df.columns)

        if attribute == 'centers':
            try:
                assert value.shape[0] == 3

                self.df.loc['values', attribute] = value

            except AssertionError:
                print('centers length must be 3: XYZ')

        else:
            self.df.loc['values', attribute] = value

        return self

    @_setdoc_pro([ds.centers, ds.rescaling_factor])
    def rescale_data(self,
                     rescaling_factor=None,
                     centers=None,
                     axial_anisotropy=None
                     ):
        """
        Rescale inplace: surface_points, orientations---adding columns in the categories_df---and grid---adding values_r
        attributes. The rescaled values will get stored on the linked objects.

        Args:
            rescaling_factor: [s1]
            centers: [s0]

        Returns:

        """

        xyz = self.concat_surface_points_orientations(self.surface_points.df[['X', 'Y', 'Z']],
                                                      self.orientations.df[['X', 'Y', 'Z']])

        # This is asking for XYZ parameters
        max_coord, min_coord = self.max_min_coord(xyz)

        if rescaling_factor is None:
            # This is asking for XYZ parameters
            self.df['rescaling factor'] = self.compute_rescaling_factor_for_0_1(max_coord=max_coord,
                                                                                min_coord=min_coord)
        else:
            self.df['rescaling factor'] = rescaling_factor
        if centers is None:
            # This is asking for XYZ parameters
            self.df.at['values', 'centers'] = self.compute_data_center(max_coord=max_coord,
                                                                       min_coord=min_coord)
        else:
            self.df.at['values', 'centers'] = centers

        self.set_rescaled_surface_points(axial_anisotropy=axial_anisotropy)
        self.set_rescaled_orientations(axial_anisotropy=axial_anisotropy)
        self.set_rescaled_grid(axial_anisotropy=axial_anisotropy)
        return True

    def compute_axial_anisotropy(self, type=None, extent=None):
        if type is None:
            type = self.axial_anisotropy_type

        if type == 'data':
            x1, y1, z1 = self.max_coord
            x0, y0, z0 = self.min_coord
        elif type == 'extent':
            if extent is None:
                extent = self.grid.regular_grid.extent

            x0, x1, y0, y1, z0, z1 = extent
        else:
            raise AttributeError('Type must be either data or extent')

        # Calculate average
        x_d = np.linalg.norm(x0-x1)
        y_d = np.linalg.norm(y0-y1)
        z_d = np.linalg.norm(z0-z1)
        mean_d = np.mean([x_d, y_d, z_d])
        return np.array([mean_d/x_d, mean_d/y_d, mean_d/z_d])

    def apply_axial_anisotropy(self, xyz, anisotropy):
        return xyz * anisotropy

    def get_rescaled_surface_points(self):
        """
        Get the rescaled coordinates. return an image of the interface and orientations categories_df with the X_r..
         columns

        Returns:
            :attr:`SurfacePoints.df[['X_c', 'Y_c', 'Z_c']]`
        """
        return self.surface_points.df[['X_c', 'Y_c', 'Z_c']]

    def get_rescaled_orientations(self):
        """
        Get the rescaled coordinates. return an image of the interface and orientations categories_df with the X_r..
         columns.

        Returns:
            :attr:`Orientations.df[['X_c', 'Y_c', 'Z_c']]`
        """
        return self.orientations.df[['X_c', 'Y_c', 'Z_c']]

    @staticmethod
    def concat_surface_points_orientations(surface_points_xyz=None, orientations_xyz=None) \
            -> pn.DataFrame:
        """
        Args:
            surface_points_xyz (:class:`pandas.DataFrame`): [s0]
            orientations_xyz (:class:`pandas.DataFrame`): [s1]
        Returns:

        """

        if surface_points_xyz is None and orientations_xyz is not None:
            df = orientations_xyz
        elif surface_points_xyz is not None and orientations_xyz is None:
            df = surface_points_xyz
        elif surface_points_xyz is not None and orientations_xyz is not None:
            df = pn.concat([orientations_xyz, surface_points_xyz], sort=False)
        else:
            raise AttributeError('You must pass at least one Data object')
        return df

    @_setdoc_pro([SurfacePoints.__doc__, Orientations.__doc__])
    def max_min_coord(self, df):
        """
        Find the maximum and minimum location of any input data in each cartesian coordinate

        Args:
            df

        Returns:
            tuple: max[XYZ], min[XYZ]
        """

        self.max_coord = df.max()[['X', 'Y', 'Z']]
        self.min_coord = df.min()[['X', 'Y', 'Z']]

        return self.max_coord, self.min_coord

    @_setdoc_pro([SurfacePoints.__doc__, Orientations.__doc__, ds.centers])
    def compute_data_center(self,
                            surface_points_xyz=None,
                            orientations_xyz=None,
                            max_coord=None, min_coord=None, inplace=True):
        """
        Calculate the center of the data once it is shifted between 0 and 1.

        Args:
            surface_points_xyz (:class:`pandas.DataFrame`): [s0]
            orientations_xyz (:class:`pandas.DataFrame`): [s1]
            max_coord (float): Max XYZ coordinates of all GeometricData
            min_coord (float): Min XYZ coordinates of all GeometricData
            inplace (bool): if True modify the self.df rescaling factor attribute

        Returns:
            np.array: [s2]
        """

        if max_coord is None or min_coord is None:
            max_coord, min_coord = self.max_min_coord(surface_points_xyz, orientations_xyz)

        # Get the centers of every axis
        centers = ((max_coord + min_coord) / 2).astype(float).values
        if inplace is True:
            self.df.at['values', 'centers'] = centers
        return centers

    @_setdoc_pro([SurfacePoints.__doc__, Orientations.__doc__, ds.rescaling_factor])
    def compute_rescaling_factor_for_0_1(self,
                                         surface_points_xyz=None,
                                         orientations_xyz=None,
                                         max_coord=None, min_coord=None,
                                         inplace=True):
        """
        Calculate the rescaling factor of the data to keep all coordinates between 0 and 1

        Args:
            surface_points_xyz (:class:`pandas.DataFrame`): [s0]
            orientations_xyz (:class:`pandas.DataFrame`): [s1]
            max_coord (float): Max XYZ coordinates of all GeometricData
            min_coord (float): Min XYZ coordinates of all GeometricData
            inplace (bool): if True modify the self.df rescaling factor attribute

        Returns:
            float: [s2]
        """

        if max_coord is None or min_coord is None:
            max_coord, min_coord = self.max_min_coord(surface_points_xyz, orientations_xyz)
        rescaling_factor_val = (2 * np.max(max_coord - min_coord))
        if inplace is True:
            self.df['rescaling factor'] = rescaling_factor_val
        return rescaling_factor_val

    @staticmethod
    @_setdoc_pro([SurfacePoints.__doc__, compute_data_center.__doc__,
                  compute_rescaling_factor_for_0_1.__doc__, ds.idx_sp])
    def rescale_surface_points(surface_points_xyz,
                               rescaling_factor,
                               centers=None,
                               idx: list = None):
        """
        Rescale inplace: surface_points. The rescaled values will get stored on the linked objects.

        Args:
            surface_points_xyz (:class:`pandas.DataFrame`): [s0]
            rescaling_factor: [s2]
            centers: [s1]
            idx (int, list of int): [s3]

        Returns:

        """

        if idx is None:
            idx = surface_points_xyz.index

        # Change the coordinates of surface_points
        new_coord_surface_points = (surface_points_xyz.loc[idx, ['X', 'Y', 'Z']] -
                                    centers) / rescaling_factor + 0.5001

        new_coord_surface_points.rename(columns={"X": "X_c", "Y": "Y_c", "Z": 'Z_c'},
                                        inplace=True)
        return new_coord_surface_points

    @_setdoc_pro(ds.idx_sp)
    def set_rescaled_surface_points(self,
                                    idx: Union[list, np.ndarray] = None,
                                    axial_anisotropy=None):
        """
        Set the rescaled coordinates into the surface_points categories_df

        Args:
            axial_anisotropy:
            idx (int, list of int): [s0]

        Returns:

        """
        if idx is None:
            idx = self.surface_points.df.index
        idx = np.atleast_1d(idx)

        if axial_anisotropy is None:
            axial_anisotropy = self.axial_anisotropy

        if axial_anisotropy is False:
            surface_points_xyz = self.surface_points.df
        else:
            axial_anisotropy_scale = self.compute_axial_anisotropy()
            surface_points_xyz = self.apply_axial_anisotropy(
                self.surface_points.df[['X', 'Y', 'Z']],
                axial_anisotropy_scale)

        self.surface_points.df.loc[idx, ['X_c', 'Y_c', 'Z_c']] = self.rescale_surface_points(
            surface_points_xyz, # This is asking for XYZ parameters
            self.df.loc['values', 'rescaling factor'],
            self.df.loc['values', 'centers'],
            idx=idx)

        return self.surface_points.df.loc[idx, ['X_c', 'Y_c', 'Z_c']]

    def rescale_data_point(self, data_points: np.ndarray, rescaling_factor=None, centers=None):
        """This method now is very similar to set_rescaled_surface_points passing an index

        Notes:
            So far is not used by any function
        """
        if rescaling_factor is None:
            rescaling_factor = self.df.loc['values', 'rescaling factor']
        if centers is None:
            centers = self.df.loc['values', 'centers']

        rescaled_data_point = (data_points - centers) / rescaling_factor + 0.5001

        return rescaled_data_point

    @staticmethod
    @_setdoc_pro([Orientations.__doc__, compute_data_center.__doc__, compute_rescaling_factor_for_0_1.__doc__, ds.idx_sp])
    def rescale_orientations(orientations_xyz, rescaling_factor, centers, idx: list = None):
        """
        Rescale inplace: surface_points. The rescaled values will get stored on the linked objects.

        Args:
            orientations_xyz (:class:`pandas.DataFrame`): [s0]
            rescaling_factor: [s2]
            centers: [s1]
            idx (int, list of int): [s3]

        Returns:

        """
        if idx is None:
            idx = orientations_xyz.index

        # Change the coordinates of orientations
        new_coord_orientations = (orientations_xyz.loc[idx, ['X', 'Y', 'Z']] -
                                  centers) / rescaling_factor + 0.5001

        new_coord_orientations.rename(columns={"X": "X_c", "Y": "Y_c", "Z": 'Z_c'}, inplace=True)

        return new_coord_orientations

    @_setdoc_pro(ds.idx_sp)
    def set_rescaled_orientations(self,
                                  idx: Union[list, np.ndarray] = None,
                                  axial_anisotropy=None
                                  ):
        """
        Set the rescaled coordinates into the surface_points categories_df

        Args:
            axial_anisotropy:
            idx (int, list of int): [s0]

        Returns:

        """
        if idx is None:
            idx = self.orientations.df.index
        idx = np.atleast_1d(idx)

        if axial_anisotropy is None:
            axial_anisotropy = self.axial_anisotropy

        if axial_anisotropy is False:
            orientations_xyz = self.orientations.df
        else:
            axial_anisotropy_scale = self.compute_axial_anisotropy()
            orientations_xyz = self.apply_axial_anisotropy(
                self.orientations.df[['X', 'Y', 'Z']],
                axial_anisotropy_scale)

        self.orientations.df.loc[idx, ['X_c', 'Y_c', 'Z_c']] = self.rescale_orientations(
            orientations_xyz,
            self.df.loc['values', 'rescaling factor'],
            self.df.loc['values', 'centers'],
            idx=idx
        )
        return self.orientations.df.loc[idx, ['X_c', 'Y_c', 'Z_c']]

    @staticmethod
    def rescale_grid(grid_extent, grid_values, rescaling_factor, centers: pn.DataFrame):
        new_grid_extent = (grid_extent - np.repeat(centers, 2)) / rescaling_factor + 0.5001
        new_grid_values = (grid_values - centers) / rescaling_factor + 0.5001
        return new_grid_extent, new_grid_values,

    def set_rescaled_grid(self, axial_anisotropy=None):
        """
        Set the rescaled coordinates and extent into a grid object
        """
        if axial_anisotropy is None:
            axial_anisotropy = self.axial_anisotropy

        # The grid has to be rescaled for having the model in scaled coordinates
        # between 0 and 1 but with the actual proportions
        self.grid.extent_r, self.grid.values_r = self.rescale_grid(
            self.grid.regular_grid.extent,
            self.grid.values,
            self.df.loc['values', 'rescaling factor'],
            self.df.loc['values', 'centers']
        )

        self.grid.regular_grid.extent_r, self.grid.regular_grid.values_r = self.grid.extent_r, self.grid.values_r

        # For the grid

        if axial_anisotropy is True:

            axial_anisotropy_scale = self.compute_axial_anisotropy()

            ani_grid_values = self.apply_axial_anisotropy(
                self.grid.values,
                axial_anisotropy_scale)

            axis_extended_l = self.apply_axial_anisotropy(
                self.grid.regular_grid.extent[[0, 2, 4]],
                axial_anisotropy_scale)

            axis_extended_r = self.apply_axial_anisotropy(
                self.grid.regular_grid.extent[[1, 3, 5]],
                axial_anisotropy_scale)

            ani_grid_extent = np.array([axis_extended_l[0],
                                        axis_extended_r[0],
                                        axis_extended_l[1],
                                        axis_extended_r[1],
                                        axis_extended_l[2],
                                        axis_extended_r[2]])

            self.grid.extent_c, self.grid.values_c = self.rescale_grid(
                ani_grid_extent,
                ani_grid_values,
                self.df.loc['values', 'rescaling factor'],
                self.df.loc['values', 'centers']
            )
        else:
            self.grid.values_c = self.grid.values_r
            self.grid.extent_c = self.grid.extent_r

        return self.grid.values_c



import gempy_legacy as gp
import pandas as pn
import numpy as np
import os
import pytest
import matplotlib.pyplot as plt
input_path = os.path.dirname(__file__)+'/../../notebooks/data'


class TestComplexModel:

    def test_init_model(self, model_complex):
        print(model_complex)

    def test_get_data(self, model_complex):
        print(gp.get_data(model_complex))
        print(gp.get_data(model_complex, itype='additional_data'))



from os import path
import sys

# This is for sphenix to find the packages
sys.path.append(path.dirname( path.dirname( path.abspath(__file__) ) ) )

from gempy_legacy.core.model import *
from gempy_legacy.utils.meta import _setdoc

# This warning comes from numpy complaining about a aesara optimization
warnings.filterwarnings("ignore",
                        message='.* a non-tuple sequence for multidimensional indexing is deprecated; use*.',
                        append=True)


@_setdoc(Series.__doc__) 
def create_series(series_distribution=None, order=None):
    return Series(series_distribution=series_distribution, order=order)


@_setdoc(Surfaces.__doc__)
def create_formations(values_array=None, values_names=np.empty(0), formation_names=np.empty(0)):
    f = Surfaces(values_array=values_array, properties_names=values_names, formation_names=formation_names)
    return f


@_setdoc(Faults.__doc__)
def create_faults(series: Series, series_fault=None, rel_matrix=None):
    return Faults(series=series, series_fault=series_fault, rel_matrix=rel_matrix)


@_setdoc(Grid.__doc__)
def create_grid(grid_type: str, **kwargs):
    return Grid(grid_type=grid_type, **kwargs)


import warnings
from typing import Union

import pandas as pn
import sys

import numpy as np

from ..surfaces import Surfaces
from .surface_points import SurfacePoints
from ..checkers import check_for_nans
from .geometric_data import GeometricData
from ...utils import docstring as ds
from ...utils.meta import _setdoc_pro


@_setdoc_pro([Surfaces.__doc__, ds.coord_ori, ds.surface_sp, ds.pole_vector, ds.orientations])
class Orientations(GeometricData):
    """
    Data child with specific methods to manipulate orientation data. It is initialize without arguments to give
    flexibility to the origin of the data.

    Args:
        surfaces (:class:`Surfaces`): [s0]
        coord (np.ndarray): [s1]
        pole_vector (np.ndarray): [s3]
        orientation (np.ndarray): [s4]
        surface (list[str]): [s2]
    Attributes:
        df (:class:`pn.core.frame.DataFrames`): Pandas data frame containing the necessary information respect
         the orientations of the model
    """

    def __init__(self, surfaces: Surfaces, coord=None, pole_vector=None, orientation=None, surface=None):
        super().__init__(surfaces)
        self._columns_o_all = ['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', 'dip', 'azimuth', 'polarity',
                               'surface', 'series', 'id', 'order_series', 'surface_number']
        self._columns_o_1 = ['X', 'Y', 'Z', 'X_c', 'Y_c', 'Z_c', 'G_x', 'G_y', 'G_z', 'dip', 'azimuth', 'polarity',
                             'surface', 'series', 'id', 'order_series', 'isFault']
        self._columns_o_num = ['X', 'Y', 'Z', 'X_c', 'Y_c', 'Z_c', 'G_x', 'G_y', 'G_z', 'dip', 'azimuth', 'polarity']
        self._columns_rend = ['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', 'smooth', 'surface']

        if (np.array(sys.version_info[:2]) <= np.array([3, 6])).all():
            self.df: pn.DataFrame

        self.set_orientations(coord, pole_vector, orientation, surface)

    @_setdoc_pro([ds.coord_ori, ds.surface_sp, ds.pole_vector, ds.orientations])
    def set_orientations(self, coord: np.ndarray = None, pole_vector: np.ndarray = None,
                         orientation: np.ndarray = None, surface: list = None):
        """
        Set coordinates, surface and orientation data.

        If both are passed pole vector has priority over orientation

        Args:
            coord (np.ndarray): [s0]
            pole_vector (np.ndarray): [s2]
            orientation (np.ndarray): [s3]
            surface (list[str]): [s1]

        Returns:

        """
        self.df = pn.DataFrame(columns=['X', 'Y', 'Z', 'X_c', 'Y_c', 'Z_c', 'G_x', 'G_y', 'G_z', 'dip',
                                        'azimuth', 'polarity', 'surface'], dtype=float)

        self.df['surface'] = self.df['surface'].astype('category', copy=True)
        self.df['surface'] = self.df['surface'].cat.set_categories(self.surfaces.df['surface'].values)

        pole_vector = check_for_nans(pole_vector)
        orientation = check_for_nans(orientation)

        if coord is not None and ((pole_vector is not None) or (orientation is not None)) and surface is not None:

            self.df[['X', 'Y', 'Z']] = pn.DataFrame(coord)
            self.df['surface'] = surface
            if pole_vector is not None:
                self.df['G_x'] = pole_vector[:, 0]
                self.df['G_y'] = pole_vector[:, 1]
                self.df['G_z'] = pole_vector[:, 2]
                self.calculate_orientations()

                if orientation is not None:
                    warnings.warn('If pole_vector and orientation are passed pole_vector is used/')
            else:
                if orientation is not None:
                    self.df['azimuth'] = orientation[:, 0]
                    self.df['dip'] = orientation[:, 1]
                    self.df['polarity'] = orientation[:, 2]
                    self.calculate_gradient()
                else:
                    raise AttributeError('At least pole_vector or orientation should have been passed to reach'
                                         'this point. Check previous condition')

        self.df['surface'] = self.df['surface'].astype('category', copy=True)
        self.df['surface'] = self.df['surface'].cat.set_categories(self.surfaces.df['surface'].values)

        self.init_dependent_properties()

        # Add nugget effect
        self.df['smooth'] = 0.01
        assert ~self.df['surface'].isna().any(), 'Some of the surface passed does not exist in the Formation' \
                                                 'object. %s' % self.df['surface'][self.df['surface'].isna()]

    @_setdoc_pro([ds.x, ds.y, ds.z, ds.surface_sp, ds.pole_vector, ds.orientations, ds.idx_sp])
    def add_orientation(self, x, y, z, surface: list[str] | str, pole_vector: Union[list, tuple, np.ndarray] = None,
                        orientation: Union[list, np.ndarray] = None, idx=None):
        """
        Add orientation.

        Args:
            x (float, np.ndarray): [s0]
            y (float, np.ndarray): [s1]
            z (float, np.ndarray): [s2]
            surface (list[str], str): [s3]
            pole_vector (np.ndarray): [s4]
            orientation (np.ndarray): [s5]
            idx (Optional[int, list[int]): [s6]

        Returns:
            Orientations
        """
        if pole_vector is None and orientation is None:
            raise AttributeError('Either pole_vector or orientation must have a value. If both are passed pole_vector'
                                 'has preference')

        max_idx = self.df.index.max()

        if idx is None:
            idx = max_idx
            if idx is np.nan:
                idx = 0
            else:
                idx += 1

        if max_idx is not np.nan:
            self.df.loc[idx] = self.df.loc[max_idx]

        if pole_vector is not None:
            self.df.loc[idx, ['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z']] = np.array([x, y, z, *pole_vector], dtype=float)
            self.calculate_orientations(idx)

            if orientation is not None:
                warnings.warn('If pole_vector and orientation are passed pole_vector is used/')
        elif orientation is not None:
            self.df.loc[idx, ['X', 'Y', 'Z', ]] = np.array([x, y, z], dtype=float)
            self.df.loc[idx, ['azimuth', 'dip', 'polarity']] = np.array(orientation, dtype=float)
            self.calculate_gradient(idx)
        else:
            raise AttributeError('At least pole_vector or orientation should have been passed to reach'
                                 'this point. Check previous condition')
        
        self._add_surface_to_list_from_new_surface_points_or_orientations(idx=idx, surface=surface)
        # if type(idx) is int:
        #     self.df.loc[idx, 'surface'] = surface[0]
        # elif type(idx) is list:
        #     self.df.loc[idx, 'surface'] = surface
        
        self.df.loc[idx, ['smooth']] = 0.01
        
        # create new pandas categories from slef.df.['surface']
        self.df['surface'] = self.df['surface'].astype('category', copy=True)
        self.df['surface'] = self.df['surface'].cat.set_categories(self.surfaces.df['surface'].values)

        self.df['series'] = self.df['series'].astype('category', copy=True)
        self.df['series'] = self.df['series'].cat.set_categories(self.surfaces.df['series'].cat.categories)

        self.map_data_from_surfaces(self.surfaces, 'series', idx=idx)
        self.map_data_from_surfaces(self.surfaces, 'id', idx=idx)
        self.map_data_from_series(self.surfaces.series, 'order_series', idx=idx)

        self.sort_table()
        return self

    @_setdoc_pro()
    def del_orientation(self, idx):
        """Delete orientation

        Args:
            idx: [s_idx_sp]

        Returns:
            :class:`gempy.core.data_modules.geometric_data.Orientations`

        """
        self.df.drop(idx, inplace=True)

    @_setdoc_pro([ds.idx_sp, ds.surface_sp])
    def modify_orientations(self, idx, **kwargs):
        """Allows modification of any of an orientation column at a given index.

        Args:
            idx (int, list[int]): [s0]

        Keyword Args:
                * X
                * Y
                * Z
                * G_x
                * G_y
                * G_z
                * dip
                * azimuth
                * polarity
                * surface (str): [s1]

         Returns:
            :class:`gempy.core.data_modules.geometric_data.Orientations`

         """

        idx = np.array(idx, ndmin=1)
        try:
            surface_names = kwargs.pop('surface')
            self.df.loc[idx, ['surface']] = surface_names
            self.map_data_from_surfaces(self.surfaces, 'series', idx=idx)
            self.map_data_from_surfaces(self.surfaces, 'id', idx=idx)
            self.map_data_from_series(self.surfaces.series, 'order_series', idx=idx)
            self.sort_table()
        except KeyError:
            pass

        # TODO Dep
        keys = list(kwargs.keys())

        # Check idx exist in the df
        assert np.isin(np.atleast_1d(idx), self.df.index).all(), 'Indices must exist in the dataframe to be modified.'

        # Check the properties are valid
        assert np.isin(list(kwargs.keys()), ['X', 'Y', 'Z', 'G_x', 'G_y', 'G_z', 'dip',
                                             'azimuth', 'polarity', 'surface', 'smooth']).all(),\
            'Properties must be one or more of the following: \'X\', \'Y\', \'Z\', \'G_x\', \'G_y\', \'G_z\', \'dip,\''\
            '\'azimuth\', \'polarity\', \'surface\''

        # stack properties values
        values = np.atleast_1d(list(kwargs.values()))

        # If we pass multiple index we need to transpose the numpy array
        if type(idx) is list or type(idx) is np.ndarray:
            values = values.T

        if values.shape[0] == 1:
            values = np.repeat(values, idx.shape[0])

        # Selecting the properties passed to be modified
        self.df.loc[idx, list(kwargs.keys())] = values.astype('float64')

        if np.isin(list(kwargs.keys()), ['G_x', 'G_y', 'G_z']).any():
            self.calculate_orientations(idx)
        else:
            if np.isin(list(kwargs.keys()), ['azimuth', 'dip', 'polarity']).any():
                self.calculate_gradient(idx)
        return self

    def calculate_gradient(self, idx=None):
        """
        Calculate the gradient vector of module 1 given dip and azimuth to be able to plot the orientations
        """
        if idx is None:
            self.df['G_x'] = np.sin(np.deg2rad(self.df["dip"].astype('float'))) * \
                np.sin(np.deg2rad(self.df["azimuth"].astype('float'))) * \
                self.df["polarity"].astype('float') + 1e-12
            self.df['G_y'] = np.sin(np.deg2rad(self.df["dip"].astype('float'))) * \
                np.cos(np.deg2rad(self.df["azimuth"].astype('float'))) * \
                self.df["polarity"].astype('float') + 1e-12
            self.df['G_z'] = np.cos(np.deg2rad(self.df["dip"].astype('float'))) * \
                self.df["polarity"].astype('float') + 1e-12
        else:
            self.df.loc[idx, 'G_x'] = np.sin(np.deg2rad(self.df.loc[idx, "dip"].astype('float'))) * \
                                      np.sin(np.deg2rad(self.df.loc[idx, "azimuth"].astype('float'))) * \
                                      self.df.loc[idx, "polarity"].astype('float') + 1e-12
            self.df.loc[idx, 'G_y'] = np.sin(np.deg2rad(self.df.loc[idx, "dip"].astype('float'))) * \
                np.cos(np.deg2rad(self.df.loc[idx, "azimuth"].astype('float'))) * \
                self.df.loc[idx, "polarity"].astype('float') + 1e-12
            self.df.loc[idx, 'G_z'] = np.cos(np.deg2rad(self.df.loc[idx, "dip"].astype('float'))) * \
                self.df.loc[idx, "polarity"].astype('float') + 1e-12
        return True

    def calculate_orientations(self, idx=None):
        """
        Calculate and update the orientation data (azimuth and dip) from gradients in the data frame.

        Authors: Elisa Heim, Miguel de la Varga
        """
        if idx is None:
            self.df['polarity'] = 1
            self.df["dip"] = np.rad2deg(np.nan_to_num(np.arccos(self.df["G_z"] / self.df["polarity"])))

            self.df["azimuth"] = np.rad2deg(np.nan_to_num(np.arctan2(self.df["G_x"] / self.df["polarity"],
                                                                     self.df["G_y"] / self.df["polarity"])))
            self.df["azimuth"][self.df["azimuth"] < 0] += 360  # shift values from [-pi, 0] to [pi,2*pi]
            self.df["azimuth"][self.df["dip"] < 0.001] = 0  # because if dip is zero azimuth is undefined

        else:

            self.df.loc[idx, 'polarity'] = 1
            self.df.loc[idx, "dip"] = np.rad2deg(np.nan_to_num(np.arccos(self.df.loc[idx, "G_z"] /
                                                                         self.df.loc[idx, "polarity"])))

            self.df.loc[idx, "azimuth"] = np.rad2deg(np.nan_to_num(
                np.arctan2(self.df.loc[idx, "G_x"] / self.df.loc[idx, "polarity"],
                           self.df.loc[idx, "G_y"] / self.df.loc[idx, "polarity"])))

            self.df["azimuth"][self.df["azimuth"] < 0] += 360  # shift values from [-pi, 0] to [pi,2*pi]
            self.df["azimuth"][self.df["dip"] < 0.001] = 0  # because if dip is zero azimuth is undefined

        return True

    @_setdoc_pro([SurfacePoints.__doc__])
    def create_orientation_from_surface_points(self, surface_points: SurfacePoints, indices):
        # TODO test!!!!
        """
        Create and set orientations from at least 3 points categories_df

        Args:
            surface_points (:class:`SurfacePoints`): [s0]
            indices (list[int]): indices of the surface point used to generate the orientation. At least
             3 independent points will need to be passed.
        """
        selected_points = surface_points.df[['X', 'Y', 'Z']].loc[indices].values.T

        center, normal = self.plane_fit(selected_points)
        orientation = self.get_orientation(normal)

        return np.array([*center, *orientation, *normal])

    def set_default_orientation(self):
        """
        Set a default point at the middle of the extent area to be able to start making the model
        """
        if self.df.shape[0] == 0:
            self.add_orientation(.00001, .00001, .00001,
                                 self.surfaces.df['surface'].iloc[0],
                                 [0, 0, 1],
                                 )

    @_setdoc_pro([ds.file_path, ds.debug, ds.inplace])
    def read_orientations(self, table_source, debug=False, inplace=True, kwargs_pandas: dict = None, **kwargs):
        """
        Read tabular using pandas tools and if inplace set it properly to the surface points object.

        Args:
            table_source (str, path object, file-like object, or direct data frame): [s0]
            debug (bool): [s1]
            inplace (bool): [s2]
            kwargs_pandas: kwargs for the panda function :func:`pn.read_csv`
            **kwargs:
                * update_surfaces (bool): If True add to the linked `Surfaces` object unique surface names read on
                  the csv file
                * coord_x_name (str): Name of the header on the csv for this attribute, e.g for coord_x. Default X
                * coord_y_name (str): Name of the header on the csv for this attribute. Default Y
                * coord_z_name (str): Name of the header on the csv for this attribute. Default Z
                * coord_x_name (str): Name of the header on the csv for this attribute. Default G_x
                * coord_y_name (str): Name of the header on the csv for this attribute. Default G_y
                * coord_z_name (str): Name of the header on the csv for this attribute. Default G_z
                * azimuth_name (str): Name of the header on the csv for this attribute. Default azimuth
                * dip_name     (str): Name of the header on the csv for this attribute. Default dip
                * polarity_name (str): Name of the header on the csv for this attribute. Default polarity
                * surface_name (str): Name of the header on the csv for this attribute. Default formation


        Returns:

        See Also:
            :meth:`GeometricData.read_data`
        """
        coord_x_name = kwargs.get('coord_x_name', "X")
        coord_y_name = kwargs.get('coord_y_name', "Y")
        coord_z_name = kwargs.get('coord_z_name', "Z")
        g_x_name = kwargs.get('G_x_name', 'G_x')
        g_y_name = kwargs.get('G_y_name', 'G_y')
        g_z_name = kwargs.get('G_z_name', 'G_z')
        azimuth_name = kwargs.get('azimuth_name', 'azimuth')
        dip_name = kwargs.get('dip_name', 'dip')
        polarity_name = kwargs.get('polarity_name', 'polarity')
        surface_name = kwargs.get('surface_name', "formation")

        if kwargs_pandas is None:
            kwargs_pandas = {}

        if 'sep' not in kwargs_pandas:
            kwargs_pandas['sep'] = ','

        if isinstance(table_source, pn.DataFrame):
            table = table_source
        else:
            table = pn.read_csv(table_source, **kwargs_pandas)

        if 'update_surfaces' in kwargs:
            if kwargs['update_surfaces'] is True:
                self.surfaces.add_surface(table[surface_name].unique())

        if debug is True:
            print('Debugging activated. Changes won\'t be saved.')
            return table

        else:
            assert np.logical_or({coord_x_name, coord_y_name, coord_z_name, dip_name, azimuth_name,
                    polarity_name, surface_name}.issubset(table.columns),
                 {coord_x_name, coord_y_name, coord_z_name, g_x_name, g_y_name, g_z_name,
                  polarity_name, surface_name}.issubset(table.columns)), \
                "One or more columns do not match with the expected values, which are: \n" +\
                "- the locations of the measurement points '{}','{}' and '{}' \n".format(coord_x_name,coord_y_name,
                                                                                         coord_z_name)+ \
                "- EITHER '{}' (trend direction indicated by an angle between 0 and 360 with North at 0 AND " \
                "'{}' (inclination angle, measured from horizontal plane downwards, between 0 and 90 degrees) \n".format(
                azimuth_name, dip_name) +\
                "- OR the pole vectors of the orientation in a cartesian system '{}','{}' and '{}' \n".format(g_x_name,
                                                                                                              g_y_name,
                                                                                                              g_z_name)+\
                "- the '{}' of the orientation, can be normal (1) or reversed (-1) \n".format(polarity_name)+\
                "- the name of the surface: '{}'\n".format(surface_name)+\
                "Your headers are "+str(list(table.columns))

            if inplace:
                # self.categories_df[table.columns] = table
                c = np.array(self._columns_o_1)
                orientations_read = table.assign(**dict.fromkeys(c[~np.in1d(c, table.columns)], np.nan))
                self.set_orientations(coord=orientations_read[[coord_x_name, coord_y_name, coord_z_name]],
                                      pole_vector=orientations_read[[g_x_name, g_y_name, g_z_name]].values,
                                      orientation=orientations_read[[azimuth_name, dip_name, polarity_name]].values,
                                      surface=orientations_read[surface_name])
            else:
                return table

    def update_annotations(self):
        """
        Add a column in the Dataframes with latex names for each input_data paramenter.

        Returns:

        """
        orientation_num = self.df.groupby('id').cumcount()
        foli_l = [r'${\bf{x}}_{\beta \,{\bf{' + str(f) + '}},' + str(p) + '}$'
                  for p, f in zip(orientation_num, self.df['id'])]

        self.df['annotations'] = foli_l
        return self

    @staticmethod
    def get_orientation(normal):
        """Get orientation (dip, azimuth, polarity ) for points in all point set"""

        # calculate dip
        dip = np.arccos(normal[2]) / np.pi * 180.

        # calculate dip direction
        # +/+
        if normal[0] >= 0 and normal[1] > 0:
            dip_direction = np.arctan(normal[0] / normal[1]) / np.pi * 180.
        # border cases where arctan not defined:
        elif normal[0] > 0 and normal[1] == 0:
            dip_direction = 90
        elif normal[0] < 0 and normal[1] == 0:
            dip_direction = 270
        # +-/-
        elif normal[1] < 0:
            dip_direction = 180 + np.arctan(normal[0] / normal[1]) / np.pi * 180.
        # -/-
        elif normal[0] < 0 >= normal[1]:
            dip_direction = 360 + np.arctan(normal[0] / normal[1]) / np.pi * 180.
        # if dip is just straight up vertical
        elif normal[0] == 0 and normal[1] == 0:
            dip_direction = 0

        else:
            raise ValueError('The values of normal are not valid.')

        if -90 < dip < 90:
            polarity = 1
        else:
            polarity = -1

        return dip, dip_direction, polarity

    @staticmethod
    def plane_fit(point_list):
        """
        Fit plane to points in PointSet
        Fit an d-dimensional plane to the points in a point set.
        adjusted from: http://stackoverflow.com/questions/12299540/plane-fitting-to-4-or-more-xyz-points

        Args:
            point_list (array_like): array of points XYZ

        Returns:
            Return a point, p, on the plane (the point-cloud centroid),
            and the normal, n.
        """

        points = point_list

        from numpy.linalg import svd
        points = np.reshape(points, (np.shape(points)[0], -1))  # Collapse trialing dimensions
        assert points.shape[0] <= points.shape[1], "There are only {} points in {} dimensions.".format(points.shape[1],
                                                                                                       points.shape[0])
        ctr = points.mean(axis=1)
        x = points - ctr[:, np.newaxis]
        M = np.dot(x, x.T)  # Could also use np.cov(x) here.

        # ctr = Point(x=ctr[0], y=ctr[1], z=ctr[2], type='utm', zone=self.points[0].zone)
        normal = svd(M)[0][:, -1]
        # return ctr, svd(M)[0][:, -1]
        if normal[2] < 0:
            normal = - normal

        return ctr, normal



import warnings

import numpy as np
import pandas as pn

from .surfaces import Surfaces
from .grid import Grid
from .structure import Structure

# This is for sphenix to find the packages
from ..utils.meta import _setdoc_pro

pn.options.mode.chained_assignment = None


class Options(object):
    """The class options contains the auxiliary user editable flags mainly independent to the model.

     Attributes:
        df (:class:`pn.DataFrame`): df containing the flags. All fields are pandas categories allowing the user to
         change among those categories.

     """

    def __init__(self):
        df_ = pn.DataFrame(np.array(['float32', 'geology', 'fast_compile', 'cpu', None]).reshape(1, -1),
                           index=['values'],
                           columns=['dtype', 'output', 'aesara_optimizer', 'device', 'verbosity'])

        self.df = df_.astype({'dtype': 'category', 'output': 'category',
                              'aesara_optimizer': 'category', 'device': 'category',
                              'verbosity': object})

        self.df['dtype'] = self.df['dtype'].cat.set_categories(['float32', 'float64'])
        self.df['aesara_optimizer'] = self.df['aesara_optimizer'].cat.set_categories(['fast_run', 'fast_compile'])
        self.df['device'] = self.df['device'].cat.set_categories(['cpu', 'cuda'])

        self.default_options()

    def __repr__(self):
        return self.df.T.to_string()

    def _repr_html_(self):
        return self.df.T.to_html()

    def modify_options(self, attribute, value):
        """Method to modify a given field

        Args:
            attribute (str): Name of the field to modify
            value: new value of the field. It will have to exist in the category in order for pandas to modify it.

        Returns:
            :class:`pandas.DataFrame`: df where options data is stored
        """

        assert np.isin(attribute, self.df.columns).all(), 'Valid properties are: ' + np.array2string(self.df.columns)
        self.df.loc['values', attribute] = value
        return self.df

    def default_options(self):
        """Set default options.

        Returns:
            bool: True
        """
        import aesara
        self.df.loc['values', 'device'] = aesara.config.device

        if self.df.loc['values', 'device'] == 'cpu':
            self.df.loc['values', 'dtype'] = 'float64'
        else:
            self.df.loc['values', 'dtype'] = 'float32'

        self.df.loc['values', 'aesara_optimizer'] = 'fast_compile'
        return True


@_setdoc_pro([Grid.__doc__, Structure.__doc__])
class KrigingParameters(object):
    """
    Class that stores and computes the default values for the kriging parameters used during the interpolation.
    The default values will be computed from the :class:`Grid` and :class:`Structure` linked objects

    Attributes:
        grid (:class:`Grid`): [s0]
        structure (:class:`Structure`): [s1]
        df (:class:`pn.DataFrame`): df containing the kriging parameters.

    Args:
        grid (:class:`Grid`): [s0]
        structure (:class:`Structure`): [s1]
    """

    def __init__(self, grid: Grid, structure: Structure):
        self.structure = structure
        self.grid = grid

        df_ = pn.DataFrame(np.array([np.nan, np.nan, 3]).reshape(1, -1),
                           index=['values'],
                           columns=['range', '$C_o$', 'drift equations',
                                    ])

        self.df = df_.astype({'drift equations': object,
                              'range': object,
                              '$C_o$': object})
        self.set_default_range()
        self.set_default_c_o()
        self.set_u_grade()

    def __repr__(self):
        return self.df.T.to_string()

    def _repr_html_(self):
        return self.df.T.to_html()

    def modify_kriging_parameters(self, attribute: str, value, **kwargs):
        """Method to modify a given field

        Args:
            attribute (str): Name of the field to modify
            value: new value of the field. It will have to exist in the category in order for pandas to modify it.
            kwargs:
                * u_grade_sep (str): If drift equations values are `str`, symbol that separates the values.

        Returns:
            :class:`pandas.DataFrame`: df where options data is stored
        """

        u_grade_sep = kwargs.get('u_grade_sep', ',')

        assert np.isin(attribute, self.df.columns).all(), 'Valid properties are: ' + np.array2string(self.df.columns)

        if attribute == 'drift equations':
            value = np.asarray(value)
            print(value)

            if type(value) is str:
                value = np.fromstring(value[1:-1], sep=u_grade_sep, dtype=int)
            try:
                assert value.shape[0] is self.structure.df.loc['values', 'len series surface_points'].shape[0]
                print(value, attribute)
                self.df.at['values', attribute] = value
                print(self.df)

            except AssertionError:
                print('u_grade length must be the same as the number of series')

        else:
            self.df = self.df.astype({'drift equations': object,
                                  'range': object,
                                  '$C_o$': object})

            self.df.at['values', attribute] = value

    def str2int_u_grade(self, **kwargs):
        """
        Convert u_grade to ints

        Args:
            **kwargs:
                * u_grade_sep (str): If drift equations values are `str`, symbol that separates the values.

        Returns:

        """
        u_grade_sep = kwargs.get('u_grade_sep', ',')
        value = self.df.loc['values', 'drift equations']
        if type(value) is str:
            value = np.fromstring(value[1:-1], sep=u_grade_sep, dtype=int)
        try:
            assert value.shape[0] is self.structure.df.loc['values', 'len series surface_points'].shape[0]
            self.df.at['values', 'drift equations'] = value

        except AssertionError:
            print('u_grade length must be the same as the number of series')

        return self.df

    def set_default_range(self, extent=None):
        """
        Set default kriging_data range

        Args:
            extent (Optional[float, np.array]): extent used to compute the default range--i.e. largest diagonal. If None
             extent of the linked :class:`Grid` will be used.

        Returns:

        """
        if extent is None:
            extent = self.grid.regular_grid.extent
            if np.sum(extent) == 0 and self.grid.values.shape[0] > 1:
                extent = np.concatenate((np.min(self.grid.values, axis=0),
                                         np.max(self.grid.values, axis=0)))[[0, 3, 1, 4, 2, 5]]

        try:
            range_var = np.sqrt(
                (extent[0] - extent[1]) ** 2 +
                (extent[2] - extent[3]) ** 2 +
                (extent[4] - extent[5]) ** 2)
        except TypeError:
            warnings.warn('The extent passed or if None the extent of the grid object has some '
                          'type of problem',
                          TypeError)
            range_var = np.array(np.nan)

        self.df['range'] = np.atleast_1d(range_var)

        return range_var

    def set_default_c_o(self, range_var=None):
        """
        Set default covariance at 0.

        Args:
            range_var (Optional[float, np.array]): range used to compute the default c_0--i.e. largest diagonal. If None
             the already computed range will be used.

        Returns:

        """
        if range_var is None:
            range_var = self.df.loc['values', 'range']

        if type(range_var) is list:
            range_var = np.atleast_1d(range_var)

        self.df.at['values', '$C_o$'] = range_var ** 2 / 14 / 3

        return self.df['$C_o$']

    def set_u_grade(self, u_grade: list = None):
        """
         Set default universal grade. Transform polynomial grades to number of equations

         Args:

             u_grade (list):

         Returns:

         """
        # =========================
        # Choosing Universal drifts
        # =========================
        if u_grade is None:

            len_series_i = self.structure.df.loc['values', 'len series surface_points']
            u_grade = np.ones_like(len_series_i)
            # u_grade[(len_series_i > 1)] = 1

        else:
            u_grade = np.array(u_grade)

        # Transforming grade to number of equations
        n_universal_eq = np.zeros_like(u_grade)
        n_universal_eq[u_grade == 0] = 0
        n_universal_eq[u_grade == 1] = 3
        n_universal_eq[u_grade == 2] = 9

        self.df.at['values', 'drift equations'] = n_universal_eq
        return self.df['drift equations']


class AdditionalData(object):
    """
    Container class that encapsulate :class:`Structure`, :class:`KrigingParameters`, :class:`Options` and
     rescaling parameters

    Args:
        surface_points (:class:`SurfacePoints`): [s0]
        orientations (:class:`Orientations`): [s1]
        grid (:class:`Grid`): [s2]
        faults (:class:`Faults`): [s4]
        surfaces (:class:`Surfaces`): [s3]
        rescaling (:class:`RescaledData`): [s5]

    Attributes:
        structure_data (:class:`Structure`): [s6]
        options (:class:`Options`): [s8]
        kriging_data (:class:`Structure`): [s7]
        rescaling_data (:class:`RescaledData`):

    """

    def __init__(self, surface_points, orientations, grid: Grid,
                 faults, surfaces: Surfaces, rescaling):
        self.structure_data = Structure(surface_points, orientations, surfaces, faults)
        self.options = Options()
        self.kriging_data = KrigingParameters(grid, self.structure_data)
        self.rescaling_data = rescaling

    def __repr__(self):
        concat_ = self.get_additional_data()
        return concat_.to_string()

    def _repr_html_(self):
        concat_ = self.get_additional_data()
        return concat_.to_html()

    def get_additional_data(self):
        """
        Concatenate all linked data frames and transpose them for a nice visualization.

        Returns:
            pn.DataFrame: concatenated and transposed dataframe
        """
        concat_ = pn.concat([self.structure_data.df, self.options.df, self.kriging_data.df, self.rescaling_data.df],
                            axis=1, keys=['Structure', 'Options', 'Kriging', 'Rescaling'])
        return concat_.T

    def update_default_kriging(self):
        """
        Update default kriging values.
        """
        self.kriging_data.set_default_range()
        self.kriging_data.set_default_c_o()
        self.kriging_data.set_u_grade()

    def update_structure(self):
        """
        Update fields dependent on input data sucha as structure and universal kriging grade
        """
        self.structure_data.update_structure_from_input()
        if len(self.kriging_data.df.loc['values', 'drift equations']) < \
                self.structure_data.df.loc['values', 'number series']:
            self.kriging_data.set_u_grade()



import gempy_legacy as gp
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import numpy as np
import pytest


def test_plot_2d_data_default(one_fault_model_no_interp):
    gp.plot.plot_2d(one_fault_model_no_interp)
    plt.show()


def test_plot_2d_data_default_all_none(one_fault_model_no_interp):
    gp.plot.plot_2d(one_fault_model_no_interp, show_data=True, show_results=False)
    plt.show()


def test_basic(one_fault_model_no_interp):
    fig = plt.figure()
    ax = fig.add_subplot(111)
    ax.imshow(np.arange(25).reshape(5, 5))
    fig.show()


def test_plot_2d_test_labels(one_fault_model_no_interp):
    geo_model = one_fault_model_no_interp
    section_dict = {'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                    'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
    geo_model.set_section_grid(section_dict)
    geo_model.set_topography(fd=1.2, d_z=np.array([600, 2000]), resolution=np.array([60, 60]))

    gp.plot_2d(
        model=geo_model,
        section_names=['section_NW-SE', 'section_NW-SE', 'topography'],
        direction=['x'], cell_number=['mid'],
        show_topography=True,
        show_section_traces=True
    )
    plt.show()
    
    gp.plot_2d(geo_model,
               section_names=['section_NW-SE', 'section_NW-SE', 'topography'],
               direction=['x'], cell_number=['mid'],
               show_topography=True,
               projection_distance=100)
    plt.show()

    gp.plot_2d(geo_model,
               section_names=['section_NW-SE', 'section_NW-SE', 'topography'],
               direction=['x'], cell_number=['mid'],
               show_topography=True,
               projection_distance=1000)
    plt.show()

# gp.plot.plot_section_traces(geo_model)
    plt.show()


@pytest.fixture(scope='module')
def section_model(one_fault_model_topo_solution):
    geo_model = one_fault_model_topo_solution
    section_dict = {'section_SW-NE': ([250, 250], [1750, 1750], [100, 100]),
                    'section_NW-SE': ([250, 1750], [1750, 250], [100, 100])}
    geo_model.set_section_grid(section_dict)

    geo_model.set_active_grid('sections', reset=False)

    one_fault_model_topo_solution.update_additional_data()
    one_fault_model_topo_solution.update_to_interpolator()
    gp.compute_model(geo_model, sort_surfaces=False)
    return geo_model


def test_topo_sections_iterp2(section_model):
    # Test 1 single
    gp.plot_2d(section_model, section_names=['section_NW-SE'],
               show_topography=True)
    plt.show()

    # Test 2 plots
    gp.plot_2d(section_model, section_names=['section_NW-SE', 'section_NW-SE'],
               show_topography=True)
    plt.show()

    # Test 3 plots
    gp.plot_2d(section_model, section_names=['section_NW-SE', 'section_NW-SE', 'topography'],
               show_topography=True)
    plt.show()

    # Test 4
    gp.plot_2d(section_model, section_names=['section_NW-SE', 'section_NW-SE', 'topography'],
               direction=['x'], cell_number=['mid'],
               show_topography=True)
    plt.show()

    gp.plot.plot_section_traces(section_model)
    plt.show()


def test_show_results(section_model):
    p2d = gp.plot_2d(section_model,
                     show_data=False,
                     show_topography=False,
                     show_results=True,
                     show=True)


def test_ve(section_model):
    # Test ve
    p2d = gp.plot_2d(section_model, section_names=['section_NW-SE', 'section_NW-SE'],
                     show_topography=True, ve=3)

    plt.show()


def test_topo_resize(one_fault_model_topo_solution):
    geo_model = one_fault_model_topo_solution
    geo_model.set_topography(fd=1.2, d_z=np.array([600, 2000]), resolution=np.array([60, 60]))
    sol = gp.compute_model(geo_model, compute_mesh=False)
    gp.plot_2d(geo_model, section_names=['topography'])
    plt.show()



"""
GemPy Models - Some More Complex Examples
-----------------------------------------

"""

# %%
# Importing gempy
import gempy_legacy as gp

# Aux imports
import numpy as np
import pandas as pn
import matplotlib
import aesara

# %%
# Choose a model and load the corresponding data set in the line below
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 
# Model 1: Discordant layering
# 
# Model 2: Anticlinal layering
# 
# Model 3: Parallel steep dipping layers
# 
# Model 5: Normal faulting
# 
# Model 6: Synclinal layering
# 
# Model 7: Graben structure with two faults
# 
# Model 8: Similar to Model 5
# 
# Model 9: Parallel flat dipping layers
# 

# %%
data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'


# %%
def create_example(name_model, interpolator=None, save_pickle=False, plot_section=True):
    """
    Create an inter_data from one of the examples data_set
    
    Attr:
        name_model (str): name of the model that you want to generate. It has to be in ['Model 1' ,'Model 2', 'Model 3', 'Model 4','Model 5' 'Model 6','Model 7',
                          'Model 8', 'Model 9']
        save_pickle (bool, str): Save to a pickle the interp_data object. You can pass the path as a string otherwse
                                the default name will be given
        plot_section (bool)
    
    """
    name_list = np.array(['Model 1', 'Model 2', 'Model 3', 'Model 4', 'Model 5', 'Model 6', 'Model 7',
                          'Model 8', 'Model 9'])
    assert name_model in name_list, 'Name model must be in the following list: ' + str(name_list)
    # Extract number of the model
    n_model = name_model[-1]

    # Load right gempy geodata
    geo_data = gp.create_data(name_model, extent=[0, 2000, 0, 2000, 0, 1600], resolution=[50, 50, 50],
                              path_o=data_path + "/data/input_data/lisa_models/foliations" + n_model + ".csv",
                              path_i=data_path + "/data/input_data/lisa_models/interfaces" + n_model + ".csv")

    # Set the right sequential pile
    subset_list_1 = np.array(['Model 1'])
    subset_list_2 = np.array(['Model 5', 'Model 8'])
    subset_list_3 = np.array(['Model 2', 'Model 3', 'Model 9', 'Model 6'])
    subset_list_4 = np.array(['Model 7'])
    ### Model 1 - Discordant layering ###
    if name_model in subset_list_1:
        gp.map_series_to_surfaces(geo_data, {"Strat_Series_1": ('Sandstone', 'Siltstone', 'Shale'),
                                             "Strat_Series_2": ('Sandstone2', 'Siltstone2', 'Shale2')},
                                  )
    ### Model 5 - One normal Fault ###
    ### Model 8 - ###
    elif name_model in subset_list_2:
        gp.map_series_to_surfaces(geo_data, {"Fault_Series": 'Main_Fault',
                                             "Strat_Series": (
                                             'Sandstone', 'Siltstone', 'Shale', 'Sandstone_2', 'Schist', 'Gneiss')},
                                  )
        geo_data.set_is_fault(['Fault_Series'])
    elif name_model in subset_list_3:
        ### Model 2 - Aufwölbung (durch Salzstock?) ###
        ### Model 3+9 - Parallele NNE Schichtung ohne Verwerfung ###
        ### Model 6 - Mulde ###
        gp.map_series_to_surfaces(geo_data, {
            "Strat_Series": ('Sandstone', 'Siltstone', 'Shale', 'Sandstone_2', 'Schist', 'Gneiss')},
                                  )

    elif name_model in subset_list_4:
        ### Model 7 - Graben ###
        gp.map_series_to_surfaces(geo_data, {"Fault_1": 'Fault_1', "Fault_2": 'Fault_2',
                                             "Strat_Series": (
                                             'Sandstone', 'Siltstone', 'Shale', 'Sandstone_2', 'Schist', 'Gneiss')},
                                  )
        geo_data.set_is_fault(['Fault_1', 'Fault_2'])

    else:
        print('You would never reach this point. Look for the bug')

    # Interpolation and Computation
    if interpolator is None:
        interp_data = gp.set_interpolator(geo_data, aesara_optimizer='fast_run')
    else:
        interp_data = interpolator
        geo_data.set_aesara_function(interpolator)
    sol = gp.compute_model(geo_data)

    if plot_section is True:
        # 2D Plot
        gp.plot_2d(geo_data, cell_number=25,
                             direction='y', show_data=True)
        gp.plot_3d(geo_data, image=True)

    if save_pickle is not False:
        if type(save_pickle) is str:
            gp.save_model_to_pickle(geo_data, save_pickle)
        else:
            gp.save_model_to_pickle(geo_data, 'lisa-' + str(n_model))

    gp.save_model(geo_data)

    return interp_data


# %% 
def generate_all_models(list_of_models):
    for e, n in enumerate(list_of_models):
        if e == 0:
            interp = create_example(n, None, save_pickle=False, plot_section=True)
        else:
            create_example(n, interp, save_pickle=False, plot_section=True)


# %%
# Discordant layering 1
# =====================

create_example('Model 1')

# %%
# Discordant layering 2
# =====================

create_example('Model 2')

# %%
# Discordant layering 3
# =====================

create_example('Model 3')

# %%
# One normal Fault 1
# ==================

create_example('Model 5')

# %%
# Fold
# ====

create_example('Model 6')

# %%
# Graben
# ======

create_example('Model 7')

# %%
#  One normal Fault 2
# ===================

create_example('Model 8')

# %%
# Horizontal tilted
# =================

create_example('Model 9')


# %%

# generate_all_models(['Model 1', 'Model 2', 'Model 3',
#                     'Model 5', 'Model 6', 'Model 7',
#                     'Model 8', 'Model 9'])



import pytest
import sys, os

from gempy_legacy.addons.gempy_to_rexfile import GemPyToRex, geomodel_to_rex

import gempy_legacy

input_path = os.path.dirname(__file__)+'/../input_data'


@pytest.mark.skipif("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
                    reason="Skipping this test on Travis CI.")
class TestGemPyToRexClass:
    """Test the class that control the rexfile encoding"""

    def test_grab_mesh(self, unconformity_model_topo):
        gempy_to_rex = GemPyToRex()
        surfaces = gempy_to_rex.grab_meshes(unconformity_model_topo)
        print(surfaces)

    def test_gempymesh_to_rex(self, unconformity_model_topo):
        gempy_to_rex = GemPyToRex()
        surfaces = gempy_to_rex.grab_meshes(unconformity_model_topo)

        gempy_to_rex.gempy_meshes_to_rex(surfaces)

    def test_gempy_to_rex(self, unconformity_model_topo):
        gempy_to_rex = GemPyToRex()
        bytes = gempy_to_rex(unconformity_model_topo, app='RexView')

        print(bytes)

    def test_gempy_to_rex_old(self, unconformity_model_topo):
        bytes2 = geomodel_to_rex(unconformity_model_topo, False)

    def test_gempy_to_rex_with_topo(self):
        import pooch
        model_file = pooch.retrieve(url="https://github.com/cgre-aachen/gempy_data/raw/aesara_data/data/gempy_models/combination.zip",
                                    known_hash="e2fa667f2ad64d6027f513c4f801557681b0e318ee1e1fcf30fc1c7ed0c60faf")

        geo_model = gempy_legacy.load_model(name='combination', path=model_file)
        gempy_to_rex = GemPyToRex()
        bytes = gempy_to_rex(geo_model)
        print(bytes)


# Importing GemPy
import gempy_legacy as gp


# Importing auxiliary libraries
import numpy as np
import pandas as pn
import matplotlib.pyplot as plt
import pytest

import gempy_legacy.core.grid
import gempy_legacy.core.surfaces
import gempy_legacy.core.data_modules.geometric_data
import gempy_legacy.core.data_modules.orientations
import gempy_legacy.core.data_modules.scaling_system
import gempy_legacy.core.data_modules.stack
import gempy_legacy.core.data_modules.surface_points


@pytest.fixture(scope='module')
def create_faults():
    faults = gempy_legacy.core.data_modules.stack.Faults()
    return faults


@pytest.fixture(scope='module')
def create_series(create_faults):
    faults = create_faults

    series = gempy_legacy.core.data_modules.stack.Series(faults)
    series.set_series_index(['foo', 'foo2', 'foo5', 'foo7'])
    series.add_series('foo3')
    series.delete_series('foo2')
    series.rename_series({'foo': 'boo'})
    series.reorder_series(['foo3', 'boo', 'foo7', 'foo5'])

    faults.set_is_fault(['boo'])

    fr = np.zeros((4, 4))
    fr[2, 2] = True
    faults.set_fault_relation(fr)

    series.add_series('foo20')

    return series


@pytest.fixture(scope='module')
def create_surfaces(create_series):
    series = create_series
    surfaces = gempy_legacy.Surfaces(series)
    surfaces.set_surfaces_names(['foo', 'foo2', 'foo5'])

    print(series)

    # We can add new surfaces:
    surfaces.add_surface(['feeeee'])
    print(surfaces)

    # The column surface is also a pandas.Categories.
    # This will be important for the Data clases (SurfacePoints and Orientations)

    print(surfaces.df['surface'])

    ### Set values

    # To set the values we do it with the following method
    surfaces.set_surfaces_values([2, 2, 2, 5])

    print(surfaces)

    # #### Set values with a given name:

    # We can give specific names to the properties (i.e. density)
    surfaces.add_surfaces_values([[2, 2, 2, 6], [2, 2, 1, 8]], ['val_foo', 'val2_foo'])
    print(surfaces)

    ### Delete surfaces values
    #
    # To delete a full propery:
    surfaces.delete_surface_values(['val_foo', 'value_0'])

    # #### One of the surfaces must be set be the basement:

    surfaces.set_basement()
    print(surfaces)

    # #### Set surface values
    #
    # We can also use set values instead adding. This will delete the previous properties and add the new one

    surfaces.set_surfaces_values([[2, 2, 2, 6], [2, 2, 1, 8]], ['val_foo', 'val2_foo'])
    print(surfaces)

    # The last property is the correspondant series that each surface belong to. `series` and `surface`
    # are pandas categories. To get a overview of what this mean
    # check https://pandas.pydata.org/pandas-docs/stable/categorical.html.

    print(surfaces.df['series'])

    print(surfaces.df['surface'])

    # ### Map series to surface

    # To map a series to a surface we can do it by passing a dict:
    # If a series does not exist in the `Series` object, we rise a warning and we set those surfaces to nans

    d = {"foo7": 'foo', "booX": ('foo2', 'foo5', 'fee')}

    surfaces.map_series(d)
    surfaces.map_series({"foo7": 'foo', "boo": ('foo2', 'foo5', 'fee')})

    print(surfaces)

    # An advantage of categories is that they are order so no we can tidy the df by series and surface

    surfaces.df.sort_values(by='series', inplace=True)

    # If we change the basement:

    surfaces.set_basement()

    # Only one surface can be the basement:

    print(surfaces)

    # ### Modify surface name

    surfaces.rename_surfaces({'foo2': 'lala'})

    print(surfaces)

    surfaces.df.loc[2, 'val_foo'] = 22

    print(surfaces)

    # We can use `set_is_fault` to choose which of our series are faults:
    return surfaces


@pytest.fixture(scope='module')
def create_surface_points(create_surfaces, create_series):
    # # Data
    # #### SurfacePoints
    # These two DataFrames (df from now on) will contain the individual information of each point at an interface or
    # orientation. Some properties of this table are mapped from the *df* below.
    surfaces = create_surfaces
    surface_points = gempy_legacy.core.data_modules.surface_points.SurfacePoints(surfaces)

    print(surface_points)

    surface_points.set_surface_points(pn.DataFrame(np.random.rand(6, 3)), ['foo', 'foo5', 'lala', 'foo5', 'lala', 'feeeee'])

    print(surface_points)

    surface_points.map_data_from_surfaces(surfaces, 'series')
    print(surface_points)


    surface_points.map_data_from_surfaces(surfaces, 'id')
    print(surface_points)


    surface_points.map_data_from_series(create_series, 'order_series')
    print(surface_points)

    surface_points.sort_table()
    print(surface_points)
    return surface_points


@pytest.fixture(scope='module')
def create_orientations(create_surfaces, create_series):
    surfaces = create_surfaces

    # ### Orientations
    orientations = gempy_legacy.core.data_modules.orientations.Orientations(surfaces)

    print(orientations)

    # ### Set values passing pole vectors:

    orientations.set_orientations(np.random.rand(6, 3) * 10,
                                  np.random.rand(6, 3),
                                  surface=['foo', 'foo5', 'lala', 'foo5', 'lala', 'feeeee'])

    print(orientations)

    # ### Set values pasing orientation data: azimuth, dip, pole (dip direction)

    orientations.set_orientations(np.random.rand(6, 3) * 10,
                                  orientation=np.random.rand(6, 3) * 20,
                                  surface=['foo', 'foo5', 'lala', 'foo5', 'lala', 'feeeee'])

    print(orientations)

    # ### Mapping data from the other df
    orientations.map_data_from_surfaces(surfaces, 'series')
    print(orientations)

    orientations.map_data_from_surfaces(surfaces, 'id')
    print(orientations)

    orientations.map_data_from_series(create_series, 'order_series')
    print(orientations)

    orientations.update_annotations()
    return orientations


def test_add_orientation_with_pole(create_surfaces):
    orientations = gempy_legacy.Orientations(create_surfaces)
    orientations.add_orientation(1, 1, 1, 'foo', pole_vector=(1, 0, 1))
    orientations.add_orientation(2, 2, 2, 'foo', orientation=(0, 0, 1))
    orientations.add_orientation(1, 1, 1, 'foo', pole_vector=(.45, 0, .45))
    orientations.modify_orientations(2, G_x=1, G_z=1)

    print(orientations)


@pytest.fixture(scope='module')
def create_grid():
    # Test creating an empty list
    grid = gempy_legacy.core.grid.Grid()
    # Test set regular grid by hand
    grid.create_regular_grid([0, 2000, 0, 2000, -2000, 0], [50, 50, 50])
    return grid


@pytest.fixture(scope='module')
def create_rescaling(create_surface_points, create_orientations, create_grid):
    rescaling = gempy_legacy.core.data_modules.scaling_system.ScalingSystem(create_surface_points, create_orientations, create_grid)
    return rescaling


@pytest.fixture(scope='module')
def create_additional_data(create_surface_points, create_orientations, create_grid, create_faults,
                           create_surfaces, create_rescaling):

    ad = gp.AdditionalData(create_surface_points, create_orientations, create_grid, create_faults,
                           create_surfaces, create_rescaling)
    return ad


class TestDataManipulation:

    def test_series(self, create_series):
        return create_series

    def test_surfaces(self, create_surfaces):
        return create_surfaces

    def test_surface_points(self, create_surface_points):
        return create_surface_points

    def test_orientations(self, create_orientations):
        return create_orientations

    def test_rescaling(self, create_rescaling):
        return create_rescaling

    def test_additional_data(self, create_additional_data):
        return create_additional_data


def test_stack():
    stack = gempy_legacy.core.data_modules.stack.Stack()
    stack.set_series_index(['foo', 'foo2', 'foo5', 'foo7'])
    stack.add_series('foo3')
    stack.delete_series('foo2')
    stack.rename_series({'foo': 'boo'})
    stack.reorder_series(['foo3', 'boo', 'foo7', 'foo5'])
    stack.set_is_fault(['boo'])

    faults = stack
    faults.set_is_fault(['boo'])

    fr = np.zeros((4, 4))
    fr[2, 2] = True
    faults.set_fault_relation(fr)

    stack.add_series('foo20')
